# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


###################################################
# UI tests for various connection.py related bits #
###################################################


def testConnectionBlacklist(app):
    app.open(extra_opts=["--test-options=object-denylist=test-many-devices"])
    manager = app.topwin

    def _delete_vm(vmname):
        app.manager_vm_action(vmname, delete=True)
        delete = app.find_window("Delete")
        delete.find("Delete associated", "check box").click()
        delete.find("Delete", "push button").click()
        lib.utils.check(lambda: manager.active)

    lib.utils.check(lambda: "test-many-devices" not in app.topwin.fmt_nodes())
    _delete_vm("test-arm-kernel")
    _delete_vm("test-clone-full")
    _delete_vm("test-clone-simple")
    app.sleep(0.5)  # Give events time to register to hit full denylist path
    lib.utils.check(lambda: "test-many-devices" not in app.topwin.fmt_nodes())


def testConnectionConnCrash(app):
    app.open(extra_opts=["--test-options=conn-crash", "--test-options=short-poll"])
    manager = app.topwin

    manager.find(r"^test testdriver.xml - Not Connected", "table cell")
    lib.utils.check(lambda: manager.active)


def testConnectionFakeEvents(app):
    app.open(
        extra_opts=[
            "--test-options=fake-nodedev-event=computer",
            "--test-options=fake-agent-event=test-many-devices",
            "--test-options=short-poll",
        ]
    )
    manager = app.topwin
    app.sleep(1.2)  # needs a second to hit both nodedev/agent event paths
    lib.utils.check(lambda: manager.active)


def testConnectionOpenauth(app):
    app.open(extra_opts=["--test-options=fake-openauth"], window_name="Authentication required")

    dialog = app.root.find("Authentication required")

    def _run():
        username = dialog.find("Username:.*entry")
        password = dialog.find("Password:.*entry")
        username.click()
        username.text = "foo"
        app.rawinput.pressKey("Enter")
        lib.utils.check(lambda: password.focused)
        password.typeText("bar")

    _run()
    dialog.find("OK", "push button").click()
    lib.utils.check(lambda: not dialog.showing)
    manager = app.find_window("Virtual Machine Manager")
    manager.find("^test testdriver.xml$", "table cell")

    # Disconnect and reconnect to trigger it again
    def _retrigger_connection():
        manager.click()
        app.manager_conn_disconnect("test testdriver.xml")
        manager.click()
        app.manager_conn_connect("test testdriver.xml")

    _retrigger_connection()
    dialog = app.root.find("Authentication required")
    _run()
    app.rawinput.pressKey("Enter")
    lib.utils.check(lambda: not dialog.showing)
    manager = app.find_window("Virtual Machine Manager")
    manager.find("^test testdriver.xml$", "table cell")

    _retrigger_connection()
    dialog = app.root.find("Authentication required")
    dialog.find("Cancel", "push button").click()
    lib.utils.check(lambda: not dialog.showing)
    app.click_alert_button("Unable to connect", "Close")
    manager.find("test testdriver.xml - Not Connected", "table cell")


def testConnectionSessionError(app):
    app.open(extra_opts=["--test-options=fake-session-error"])
    app.click_alert_button("Could not detect a local session", "Close")



#
# Copyright 2014 Fujitsu Limited.
# Chen Hanxiao <chenhanxiao at cn.fujitsu.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _HugepagesPage(XMLBuilder):
    """
    Class representing <memoryBacking><hugepages><page> elements
    """

    XML_NAME = "page"

    size = XMLProperty("./@size")
    unit = XMLProperty("./@unit")
    nodeset = XMLProperty("./@nodeset")


class DomainMemoryBacking(XMLBuilder):
    """
    Class for generating <memoryBacking> XML
    """

    XML_NAME = "memoryBacking"
    _XML_PROP_ORDER = ["hugepages", "nosharepages", "locked", "pages"]

    hugepages = XMLProperty("./hugepages", is_bool=True)
    nosharepages = XMLProperty("./nosharepages", is_bool=True)
    locked = XMLProperty("./locked", is_bool=True)
    discard = XMLProperty("./discard", is_bool=True)
    access_mode = XMLProperty("./access/@mode")
    source_type = XMLProperty("./source/@type")
    allocation_mode = XMLProperty("./allocation/@mode")
    allocation_threads = XMLProperty("./allocation/@threads", is_int=True)

    pages = XMLChildProperty(_HugepagesPage, relative_xpath="./hugepages")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import datetime

from . import lib


###################################
# UI tests for the 'About' dialog #
###################################


def testAbout(app):
    app.root.find("Help", "menu").click()
    app.root.find("About", "menu item").click()
    win = app.root.find_fuzzy("About", "dialog")
    label = win.find_fuzzy("Copyright", "label")

    curyear = datetime.datetime.today().strftime("%Y")
    if curyear not in label.text:
        print("Current year=%s not in about.ui dialog!" % curyear)

    win.keyCombo("<ESC>")
    lib.utils.check(lambda: win.visible is False)



#
# Copyright 2014 Fujitsu Limited.
# Chen Hanxiao <chenhanxiao at cn.fujitsu.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty


class DomainResource(XMLBuilder):
    """
    Class for generating <resource> XML
    """

    XML_NAME = "resource"
    _XML_PROP_ORDER = ["partition", "fibrechannel_appid"]

    partition = XMLProperty("./partition")
    fibrechannel_appid = XMLProperty("./fibrechannel/@appid")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import tempfile

import tests
from . import lib


def _search_permissions_decorator(fn):
    """
    Decorator to set up necessary bits to test disk permission search
    """

    def wrapper(app, *args, **kwargs):
        # Generate capabilities XML from a template, with out
        # UID/GID inserted as the intended emulator permissions
        capsfile = tests.utils.UITESTDATADIR + "/capabilities/dac-caps-template.xml"
        capsdata = open(capsfile).read() % {"UID": os.getuid(), "GID": os.getgid()}
        tmpcaps = tempfile.NamedTemporaryFile(prefix="virt-manager-uitests-caps")
        tmpcapspath = tmpcaps.name
        open(tmpcapspath, "w").write(capsdata)

        # We mock a qemu URI to trigger the permissions check
        uri = tests.utils.URIs.test_full + ",fakeuri=qemu:///system,caps=%s" % tmpcapspath

        # Create a temporary directory that we can manipulate perms
        tmpobj = tempfile.TemporaryDirectory(prefix="virtinst-test-search")
        tmpdir = tmpobj.name
        try:
            os.chmod(tmpdir, 0o000)
            fn(app, uri, tmpdir, *args, **kwargs)
        finally:
            os.chmod(tmpdir, 0o777)

    return wrapper


def _select_hw(addhw, hwname, tabname):
    addhw.find(hwname, "table cell").click()
    tab = addhw.find(tabname, None)
    lib.utils.check(lambda: tab.showing)
    return tab


def _finish(addhw, check):
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    if check:
        lib.utils.check(lambda: check.active)


def _open_addhw(app, details):
    details.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    return addhw


def _open_app(app, vmname, title=None, shutdown=False, **kwargs):
    app.open(show_console=vmname, **kwargs)
    details = app.find_details_window(title or vmname, click_details=True, shutdown=shutdown)
    return details


##############
# Test cases #
##############


def testAddControllers(app):
    """
    Add various controller configs
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Default SCSI
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "SCSI")
    _finish(addhw, check=details)

    # Virtio SCSI
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "SCSI")
    tab.combo_select("Model:", "VirtIO SCSI")
    _finish(addhw, check=details)

    # USB 2
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "USB")
    tab.combo_select("Model:", "USB 2")
    _finish(addhw, check=details)

    # USB 3
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "USB")
    tab.combo_select("Model:", "USB 3")
    # Can't add more than 1 USB controller, so finish isn't sensitive
    finish = addhw.find("Finish", "push button")
    lib.utils.check(lambda: not finish.sensitive)


def testAddCephDisk(app):
    """
    Add a disk with a ceph volume, ensure it maps correctly
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Select ceph volume for disk
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    tab.find("storage-browse", "push button").click()
    browse = app.root.find("vmm-storage-browser")
    browse.find_fuzzy("pool-rbd-ceph", "table cell").bring_on_screen().click()
    browse.find_fuzzy("some-rbd-vol", "table cell").click()
    browse.find("Choose Volume", "push button").click()
    _finish(addhw, check=details)

    # Check disk details, make sure it correctly selected volume
    details.find("IDE Disk 2", "table cell").click()
    tab = details.find("disk-tab")
    lib.utils.check(lambda: tab.showing)
    disk_path = tab.find("disk-source-path")
    lib.utils.check(lambda: "rbd-sourcename/some-rbd-vol" in disk_path.text)


def testAddDisks(app):
    """
    Add various disk configs and test storage browser
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Default disk
    tab = _select_hw(addhw, "Storage", "storage-tab")
    _finish(addhw, check=details)

    # Disk with some tweaks
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.combo_select("Bus type:", "VirtIO")
    tab.find("Advanced options", "toggle button").click_expander()
    tab.find("Shareable:", "check box").click()
    tab.find("Readonly:", "check box").click()
    tab.find("Serial:", "text").set_text("ZZZZ")
    tab.combo_select("Cache mode:", "none")
    tab.combo_select("Discard mode:", "ignore")
    # High number but we are non-sparse by default so it won't complain
    tab.find("GiB", "spin button").set_text("200000")
    _finish(addhw, check=details)

    # USB disk with removable setting
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.combo_select("Bus type:", "USB")
    tab.find("Advanced options", "toggle button").click_expander()
    tab.find("Removable:", "check box").click()
    _finish(addhw, check=details)

    # Managed storage tests
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    _finish(addhw, check=None)
    app.click_alert_button("storage path must be specified", "OK")
    tab.find("storage-browse", "push button").click()
    browse = app.root.find("vmm-storage-browser")

    # Create a vol, refresh, then delete it
    browse.find_fuzzy("pool-dir", "table cell").click()
    browse.find("vol-new", "push button").click()
    newvol = app.find_window("Add a Storage Volume")
    newname = "a-newvol"
    newvol.find("Name:", "text").set_text(newname)
    newvol.find("Finish", "push button").click()
    lib.utils.check(lambda: not newvol.showing)
    volcell = browse.find(newname, "table cell")
    lib.utils.check(lambda: volcell.selected)
    browse.find("vol-refresh", "push button").click()
    volcell = browse.find(newname, "table cell")
    lib.utils.check(lambda: volcell.selected)
    browse.find("vol-delete", "push button").click()
    app.click_alert_button("permanently delete the volume", "Yes")
    lib.utils.check(lambda: volcell.dead)

    # Test browse local
    browse.find("Browse Local", "push button").click()
    chooser = app.root.find("Locate existing storage", "file chooser")

    # use filename that is near the beginning of the file list when sorted,
    # as the row in the file dialog may become scrolled out of the view and
    # cause the test to fail
    fname = "COPYING"
    chooser.find(fname, "table cell").click()
    chooser.find("Open", "push button").click()
    lib.utils.check(lambda: not chooser.showing)
    lib.utils.check(lambda: addhw.active)
    storageent = tab.find("storage-entry")
    lib.utils.check(lambda: ("/" + fname) in storageent.text)

    # Reopen dialog, select a volume, etic
    tab.find("storage-browse", "push button").click()
    browse = app.root.find("vmm-storage-browser")

    browse.find_fuzzy("pool-logical", "table cell").click()
    browse.find("diskvol1", "table cell").click()
    browse.find("Choose Volume", "push button").click()
    lib.utils.check(lambda: "/diskvol1" in storageent.text)
    _finish(addhw, check=None)
    app.click_alert_button("already in use by", "No")
    _finish(addhw, check=None)
    app.click_alert_button("already in use by", "Yes")
    lib.utils.check(lambda: details.active)

    # choose file for floppy
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.combo_select("Device type:", "Floppy device")
    diskradio = tab.find_fuzzy("Create a disk image", "radio")
    lib.utils.check(lambda: not diskradio.sensitive)
    tab.find("storage-entry").set_text("/pool-dir/bochs-vol")
    _finish(addhw, check=details)

    # empty cdrom
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.combo_select("Device type:", "CDROM device")
    tab.combo_select("Bus type:", "SCSI")
    _finish(addhw, check=details)


@_search_permissions_decorator
def testAddDiskSearchPermsCheckbox(app, uri, tmpdir):
    """
    Test search permissions 'no' and checkbox case
    """
    app.uri = uri
    details = _open_app(app, "test-clone-simple")

    # Say 'No' but path should still work due to test driver
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo1.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=None)
    app.click_alert_button("emulator may not have", "No")
    lib.utils.check(lambda: details.active)

    # Say 'don't ask again'
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo2.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=None)
    alert = app.root.find_fuzzy("vmm dialog", "alert")
    alert.find_fuzzy("Don't ask", "check box").click()
    app.click_alert_button("emulator may not have", "No")
    lib.utils.check(lambda: details.active)

    # Confirm it doesn't ask about path again
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo3.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=details)


@_search_permissions_decorator
def testAddDiskSearchPermsSuccess(app, uri, tmpdir):
    """
    Select 'Yes' for search perms fixing
    """
    app.uri = uri
    details = _open_app(app, "test-clone-simple")

    # Say 'Yes'
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo1.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=None)
    app.click_alert_button("emulator may not have", "Yes")
    lib.utils.check(lambda: details.active)

    # Confirm it doesn't ask about path again
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo3.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=details)


@_search_permissions_decorator
def testAddDiskSearchPermsFail(app, uri, tmpdir):
    """
    Force perms fixing to fail
    """
    app.uri = uri
    details = _open_app(app, "test-clone-simple", break_setfacl=True)

    # Say 'Yes' and it should fail, then denylist the paths
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo1.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=None)
    app.click_alert_button("emulator may not have", "Yes")
    alert = app.root.find("vmm dialog", "alert")
    alert.find_fuzzy("Errors were encountered", "label")
    alert.find_fuzzy("Don't ask", "check box").click()
    alert.find_fuzzy("OK", "push button").click()
    lib.utils.check(lambda: details.active)

    # Confirm it doesn't ask about path again
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    path = tmpdir + "/foo2.img"
    tab.find("storage-entry").set_text(path)
    _finish(addhw, check=details)


def testAddNetworks(app):
    """
    Test various network configs
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Basic network + opts
    tab = _select_hw(addhw, "Network", "network-tab")
    tab.combo_select("net-source", "Virtual network 'default'")
    tab.find("MAC Address Field", "text").set_text("00:11:00:11:00:11")
    tab.combo_select("Device model:", "virtio")
    _finish(addhw, check=details)

    # Portgroup
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Network", "network-tab")
    tab.find("MAC Address Field", "text").set_text("00:11:00:11:00:CC")
    tab.combo_select("net-source", "Virtual network 'plainbridge-portgroups'")
    c = tab.find_fuzzy("Portgroup:", "combo box")
    c.click_combo_entry()
    lib.utils.check(lambda: c.find("engineering", "menu item").selected)
    c.find("engineering", "menu item").click()
    _finish(addhw, check=details)

    # Manual macvtap
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Network", "network-tab")
    tab.combo_select("net-source", "Macvtap device...")
    tab.find("Device name:", "text").set_text("macvtapfoo7")
    _finish(addhw, check=details)

    # Manual bridge. Also trigger MAC collision
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Network", "network-tab")
    tab.find("mac-address-enable", "check box").click()
    tab.combo_select("net-source", "Bridge device...")
    tab.find("Device name:", "text").set_text("zbr0")
    _finish(addhw, check=None)
    # Check MAC validation error
    app.click_alert_button("00:11:22:33:44:55", "Close")

    # Fix MAC
    tab.find("mac-address-enable", "check box").click()
    tab.find("MAC Address Field", "text").set_text("00:11:0A:11:00:11")
    _finish(addhw, check=details)


def testAddGraphics(app):
    """
    Graphics device testing
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # VNC example
    tab = _select_hw(addhw, "Graphics", "graphics-tab")
    tab.combo_select("Type:", "VNC")
    tab.combo_select("Listen type:", "Address")
    tab.combo_select("Address:", "All interfaces")
    tab.find("graphics-port-auto", "check").click()
    tab.find("graphics-port", "spin button").set_text("1234")
    tab.find("Password:", "check").click()
    passwd = tab.find_fuzzy("graphics-password", "text")
    newpass = "foobar"
    passwd.typeText(newpass)
    tab.find("Show password", "check").click()
    lib.utils.check(lambda: passwd.text == newpass)
    tab.find("Show password", "check").click()
    lib.utils.check(lambda: passwd.text != newpass)
    _finish(addhw, check=None)
    # Catch a port error
    app.click_alert_button("Port must be above 5900", "Close")
    tab.find("graphics-port", "spin button").set_text("5920")
    _finish(addhw, check=details)

    # Spice regular example
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Graphics", "graphics-tab")
    tab.combo_select("Type:", "Spice")
    _finish(addhw, check=details)

    # Spice GL example
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Graphics", "graphics-tab")
    tab.combo_select("Type:", "Spice")
    tab.combo_select("Listen type:", "None")
    tab.find("OpenGL:", "check box").click()
    tab.combo_check_default("graphics-rendernode", "0000")
    _finish(addhw, check=details)


def testAddHosts(app):
    """
    Add a few different USB and PCI devices
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Add USB device dup1
    tab = _select_hw(addhw, "USB Host Device", "host-tab")
    tab.find_fuzzy("HP Dup USB 1", "table cell").click()
    _finish(addhw, check=None)
    app.click_alert_button("device is already in use by", "No")
    _finish(addhw, check=None)
    app.click_alert_button("device is already in use by", "Yes")
    lib.utils.check(lambda: details.active)

    # Add USB device dup2
    _open_addhw(app, details)
    tab = _select_hw(addhw, "USB Host Device", "host-tab")
    tab.find_fuzzy("HP Dup USB 2", "table cell").click()
    _finish(addhw, check=None)
    app.click_alert_button("device is already in use by", "Yes")
    lib.utils.check(lambda: details.active)

    # Add another USB device
    _open_addhw(app, details)
    tab = _select_hw(addhw, "USB Host Device", "host-tab")
    tab.find_fuzzy("Cruzer Micro 256", "table cell").click()
    _finish(addhw, check=details)

    # Add PCI device
    _open_addhw(app, details)
    tab = _select_hw(addhw, "PCI Host Device", "host-tab")
    tab.find_fuzzy("(Interface eth0)", "table cell").click()
    _finish(addhw, check=None)
    app.click_alert_button("device is already in use by", "Yes")
    lib.utils.check(lambda: details.active)

    # Add MDEV device
    _open_addhw(app, details)
    tab = _select_hw(addhw, "MDEV Host Device", "host-tab")
    tab.find_fuzzy("mdev_8e37ee90_2b51_45e3_9b25_bf8283c03110", "table cell").click()
    _finish(addhw, check=details)


def testAddChars(app):
    """
    Add a bunch of char devices
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Add console device
    tab = _select_hw(addhw, "Console", "char-tab")
    tab.combo_select("Device Type:", "Pseudo TTY")
    tab.combo_select("Type:", "Hypervisor default")
    _finish(addhw, check=details)

    # Add serial+file
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Serial", "char-tab")
    tab.combo_select("Device Type:", "Output to a file")
    tab.find("Path:", "text").set_text("/tmp/foo.log")
    _finish(addhw, check=details)

    _open_addhw(app, details)
    tab = _select_hw(addhw, "Parallel", "char-tab")
    tab.combo_select("Device Type:", "UNIX")
    _finish(addhw, check=details)

    # Add spicevmc channel
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Channel", "char-tab")
    tab.combo_check_default("char-target-name", ".*redhat.spice.0.*")
    tab.combo_select("char-target-name", ".*webdav.*")
    tab.combo_select("char-target-name", ".*org.qemu.guest_agent*")
    _finish(addhw, check=details)


def testAddLXCFilesystem(app):
    """
    Adding LXC specific filesystems
    """
    app.uri = tests.utils.URIs.lxc

    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Add File+nbd share
    tab = _select_hw(addhw, "Filesystem", "filesystem-tab")
    tab.combo_select("Type:", "file")
    tab.combo_select("Driver:", "nbd")
    tab.combo_select("Format:", "qcow2")

    source = tab.find("Source path:", "text")
    source.set_text("/foo/source")
    tab.find("Browse...", "push button").click()
    # Specific testing for dir vol handling for filesystem browse
    browsewin = app.root.find("vmm-storage-browser")
    browsewin.find_fuzzy("pool-dir", "table cell").click()
    browsewin.find_fuzzy("bochs-vol", "table cell").click()
    choose = browsewin.find("Choose Volume")
    lib.utils.check(lambda: not choose.sensitive)
    browsewin.find_fuzzy("dir-vol", "table cell").click()
    lib.utils.check(lambda: choose.sensitive)
    choose.click()
    lib.utils.check(lambda: addhw.active)
    lib.utils.check(lambda: source.text == "/pool-dir/dir-vol")

    tab.find_fuzzy("Export filesystem", "check").click()
    # Use this to test some error.py logic for truncating large errors
    badtarget = "a" * 1024
    tab.find("Target path:", "text").set_text(badtarget)
    _finish(addhw, check=None)
    app.click_alert_button("aaa...", "Close")
    tab.find("Target path:", "text").set_text("/foo/target")
    _finish(addhw, check=details)

    # Add RAM type
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Filesystem", "filesystem-tab")
    tab.combo_select("Type:", "ram")
    tab.find("Usage:", "spin button").set_text("12345")
    tab.find("Target path:", "text").set_text("/mem")
    _finish(addhw, check=details)


def testAddHWMisc1(app):
    """
    Add some simple devices
    """
    details = _open_app(app, "test-clone-simple", keyfile="rawdefault.ini")

    # Disk, verify that raw will fully allocate by default
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Storage", "storage-tab")
    # Size too big
    tab.find("GiB", "spin button").set_text("200000")
    _finish(addhw, check=None)
    app.click_alert_button("not enough free space", "Close")
    tab.find("GiB", "spin button").set_text("1.5")
    _finish(addhw, check=details)

    # Add input
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Input", "input-tab")
    tab.combo_select("Type:", "EvTouch")
    _finish(addhw, check=details)

    # Add sound
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Sound", "sound-tab")
    tab.combo_select("Model:", "HDA")
    _finish(addhw, check=details)

    # Add video
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Video", "video-tab")
    tab.combo_select("Model:", "Virtio")
    _finish(addhw, check=details)

    # Add watchdog
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Watchdog", "watchdog-tab")
    tab.combo_select("Model:", "I6300")
    tab.combo_select("Action:", "Pause the guest")
    _finish(addhw, check=details)

    # Add smartcard
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Smartcard", "smartcard-tab")
    tab.combo_select("Mode:", "Passthrough")
    _finish(addhw, check=details)

    # Add TPM default
    _open_addhw(app, details)
    tab = _select_hw(addhw, "TPM", "tpm-tab")
    _finish(addhw, check=details)


def testAddHWMisc2(app):
    """
    Add some more simple devices"
    """
    details = _open_app(app, "test-clone-simple")
    addhw = _open_addhw(app, details)

    # Add usb controller, to make usbredir work
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "USB")
    _finish(addhw, check=details)

    # Add usb redir
    _open_addhw(app, details)
    tab = _select_hw(addhw, "USB Redirection", "usbredir-tab")
    tab.combo_select("Type:", "Spice")
    _finish(addhw, check=details)

    # Add basic filesystem
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Filesystem", "filesystem-tab")
    tab.find("Source path:", "text").set_text("/foo/source")
    tab.find("Target path:", "text").set_text("/foo/target")
    _finish(addhw, check=details)

    # Add TPM passthrough
    _open_addhw(app, details)
    tab = _select_hw(addhw, "TPM", "tpm-tab")
    tab.combo_select("Type:", "Passthrough")
    tab.find("Device Path:", "text").set_text("/tmp/foo")
    tab.find("Advanced options", "toggle button").click_expander()
    tab.combo_select("Model:", "TIS")
    _finish(addhw, check=details)

    # Add RNG
    _open_addhw(app, details)
    tab = _select_hw(addhw, "RNG", "rng-tab")
    tab.find("Host Device:", "text").set_text("/dev/random")
    _finish(addhw, check=details)

    # Add Panic
    _open_addhw(app, details)
    tab = _select_hw(addhw, "Panic", "panic-tab")
    _finish(addhw, check=details)

    # Add vsock
    _open_addhw(app, details)
    tab = _select_hw(addhw, "VirtIO VSOCK", "vsock-tab")
    tab.find("vsock-auto").click()
    tab.find("vsock-cid").set_text("7")
    _finish(addhw, check=details)


def testAddHWUSBNone(app):
    """
    Test some special case handling when VM has controller usb model='none'
    """
    details = _open_app(
        app, "test-alternate-devs", title="test alternate devs title", shutdown=True
    )
    addhw = _open_addhw(app, details)

    # Add usb controller
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    tab.combo_select("Type:", "USB")
    _finish(addhw, check=details)

    # Trigger a libvirt error to test error handling
    addhw = _open_addhw(app, details)
    tab = _select_hw(addhw, "Controller", "controller-tab")
    combo = tab.find("Type:", "combo box")
    combo.find(None, "text").set_text("foobar")
    _finish(addhw, check=None)
    app.click_alert_button("Unable to add device", "Close")
    lib.utils.check(lambda: addhw.active)


def testAddHWCornerCases(app):
    """
    Random addhardware related tests
    """
    details = _open_app(app, "test-many-devices")
    addhw = _open_addhw(app, details)

    # Test cancel
    addhw.find("Cancel", "push button").click()

    # Test live adding, error dialog, click no
    _open_addhw(app, details)
    _finish(addhw, check=None)
    alert = app.root.find("vmm dialog", "alert")
    alert.find("This device could not be attached to the running machine", "label")
    alert.find("Details", "toggle button").click_expander()
    alert.find("No", "push button").click()
    lib.utils.check(lambda: details.active)

    # Test live adding, error dialog, click yes
    _open_addhw(app, details)
    _finish(addhw, check=None)
    alert = app.root.find("vmm dialog", "alert")
    alert.find("This device could not be attached to the running machine", "label")
    alert.find("Details", "toggle button").click_expander()
    alert.find("Yes", "push button").click()
    lib.utils.check(lambda: alert.dead)


def testAddHWXMLEdit(app):
    """
    Test XML editor integration
    """
    details = _open_app(app, "test-clone-simple", xmleditor_enabled=True)
    win = _open_addhw(app, details)

    # Disk test, change path and make sure we error it is missing
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    origpath = "/var/lib/libvirt/images/test-clone-simple.qcow2"
    newpath = "/FOO/XMLEDIT/test1.img"
    xmleditor.set_text(xmleditor.text.replace(origpath, newpath))
    _finish(win, check=None)
    app.click_alert_button("non-existent path", "Close")

    # Undo the bad change, change bus/target
    xmleditor.set_text(xmleditor.text.replace(newpath, origpath))
    xmleditor.set_text(xmleditor.text.replace("hdb", "xvda"))
    xmleditor.set_text(xmleditor.text.replace("ide", "xen"))
    _finish(win, check=details)

    # Verify the changes applied
    details.find("Xen Disk 1").click()
    lib.utils.check(lambda: details.active)
    win = _open_addhw(app, details)
    tab = _select_hw(win, "Storage", "storage-tab")
    tab.find_fuzzy("Select or create", "radio").click()
    tab.find("storage-browse", "push button").click()
    browse = app.root.find("vmm-storage-browser")
    browse.find(os.path.basename(origpath))
    browse.find("Cancel").click()

    # Select XML, switch to new dev type, verify we change focus
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    lib.utils.check(lambda: xmleditor.showing)
    tab = _select_hw(win, "Network", "network-tab")
    lib.utils.check(lambda: not xmleditor.showing)

    # Do standard xmleditor tests
    finish = win.find("Finish", "push button")
    lib.utils.test_xmleditor_interactions(app, win, finish)
    win.find("Cancel", "push button").click()
    lib.utils.check(lambda: not win.visible)



#
# Copyright 2011, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceFilesystem(Device):
    XML_NAME = "filesystem"
    _XML_PROP_ORDER = ["_type_prop", "accessmode", "fmode", "dmode"]

    TYPE_MOUNT = "mount"
    TYPE_TEMPLATE = "template"
    TYPE_FILE = "file"
    TYPE_BLOCK = "block"
    TYPE_RAM = "ram"

    MODE_MAPPED = "mapped"
    MODE_SQUASH = "squash"

    DRIVER_LOOP = "loop"
    DRIVER_NBD = "nbd"

    _type_prop = XMLProperty("./@type")
    accessmode = XMLProperty("./@accessmode")
    model = XMLProperty("./@model")
    multidevs = XMLProperty("./@multidevs")
    fmode = XMLProperty("./@fmode")
    dmode = XMLProperty("./@dmode")

    readonly = XMLProperty("./readonly", is_bool=True)
    space_hard_limit = XMLProperty("./space_hard_limit")
    space_soft_limit = XMLProperty("./space_soft_limit")

    driver_wrpolicy = XMLProperty("./driver/@wrpolicy")
    driver_type = XMLProperty("./driver/@type")
    driver_format = XMLProperty("./driver/@format")
    driver_queue = XMLProperty("./driver/@queue")
    driver_name = XMLProperty("./driver/@name")

    target_dir = XMLProperty("./target/@dir")

    source_dir = XMLProperty("./source/@dir")
    source_name = XMLProperty("./source/@name")
    source_file = XMLProperty("./source/@file")
    source_dev = XMLProperty("./source/@dev")
    source_usage = XMLProperty("./source/@usage")
    source_units = XMLProperty("./source/@units")
    source_pool = XMLProperty("./source/@pool")
    source_volume = XMLProperty("./source/@volume")
    source_socket = XMLProperty("./source/@socket")

    binary_path = XMLProperty("./binary/@path")
    binary_xattr = XMLProperty("./binary/@xattr", is_onoff=True)
    binary_cache_mode = XMLProperty("./binary/cache/@mode")
    binary_lock_posix = XMLProperty("./binary/lock/@posix", is_onoff=True)
    binary_lock_flock = XMLProperty("./binary/lock/@flock", is_onoff=True)
    binary_sandbox_mode = XMLProperty("./binary/sandbox/@mode")

    def _type_to_source_prop(self):
        if self.type == DeviceFilesystem.TYPE_TEMPLATE:
            return "source_name"
        elif self.type == DeviceFilesystem.TYPE_FILE:
            return "source_file"
        elif self.type == DeviceFilesystem.TYPE_BLOCK:
            return "source_dev"
        elif self.type == DeviceFilesystem.TYPE_RAM:
            return "source_usage"
        else:
            return "source_dir"

    def _get_source(self):
        return getattr(self, self._type_to_source_prop())

    def _set_source(self, val):
        return setattr(self, self._type_to_source_prop(), val)

    source = property(_get_source, _set_source)

    def _get_target(self):
        return self.target_dir

    def _set_target(self, val):
        self.target_dir = val

    target = property(_get_target, _set_target)

    def _get_type(self):
        return getattr(self, "_type_prop")

    def _set_type(self, val):
        # Get type/value of the attribute of "source" property
        old_source_type = self._type_to_source_prop()
        old_source_value = self.source

        # Update "type" property
        new_type = setattr(self, "_type_prop", val)

        # If the attribute type of 'source' property has changed
        # restore the value
        if old_source_type != self._type_to_source_prop():
            self.source = old_source_value

        return new_type

    type = property(_get_type, _set_type)

    ##############
    # Validation #
    ##############

    def validate_target(self, target):
        # In case of qemu for default fs type (mount) target is not
        # actually a directory, it is merely a arbitrary string tag
        # that is exported to the guest as a hint for where to mount
        if (self.conn.is_qemu() or self.conn.is_test()) and (
            self.type is None or self.type == self.TYPE_MOUNT
        ):
            return

        if not os.path.isabs(target):
            raise ValueError(_("Filesystem target '%s' must be an absolute path") % target)

    def validate(self):
        if self.target:
            self.validate_target(self.target)

    def default_accessmode(self):
        if self.driver_type == "virtiofs":
            # let libvirt fill in default accessmode=passthrough
            return None
        # libvirt qemu defaults to accessmode=passthrough, but that
        # really only works well for qemu running as root, which is
        # not the common case. so use mode=mapped
        return self.MODE_MAPPED

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        ignore = guest

        if not (self.conn.is_qemu() or self.conn.is_lxc() or self.conn.is_test()):
            return

        # type=mount is the libvirt default. But hardcode it since other
        # bits like validation depend on it
        if self.type is None:
            self.type = self.TYPE_MOUNT

        if self.accessmode is None:
            self.accessmode = self.default_accessmode()



# Copyright (C) 2006, 2013, 2014, 2015 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import threading
import time
import traceback

import libvirt

import virtinst
from virtinst import log
from virtinst import pollhelpers

from .lib import connectauth
from .lib import testmock
from .baseclass import vmmGObject
from .lib.libvirtenummap import LibvirtEnumMap
from .object.domain import vmmDomain
from .object.network import vmmNetwork
from .object.nodedev import vmmNodeDevice
from .object.storagepool import vmmStoragePool
from .lib.statsmanager import vmmStatsManager


class _ObjectList(vmmGObject):
    """
    Class that wraps our internal list of libvirt objects
    """

    # pylint: disable=not-context-manager
    # pylint doesn't know that lock() has 'with' support
    BLACKLIST_COUNT = 3

    def __init__(self):
        vmmGObject.__init__(self)

        self._objects = []
        self._denylist = {}
        self._lock = threading.Lock()

    def _cleanup(self):
        self._objects = []

    def _denylist_key(self, obj):
        return str(obj.__class__) + obj.get_name()

    def add_denylist(self, obj):
        """
        Add an object to the denylist. Basically a list of objects we
        choose not to poll, because they threw an error at init time

        :param obj: vmmLibvirtObject to denylist
        :returns: number of added object to list
        """
        key = self._denylist_key(obj)
        count = self._denylist.get(key, 0)
        self._denylist[key] = count + 1
        return self._denylist[key]

    def remove_denylist(self, obj):
        """
        :param obj: vmmLibvirtObject to remove from denylist
        :returns: True if object was denylisted or False otherwise.
        """
        key = self._denylist_key(obj)
        return bool(self._denylist.pop(key, 0))

    def in_denylist(self, obj):
        """
        If an object is in list only once don't consider it denylisted,
        give it one more chance.

        :param obj: vmmLibvirtObject to check
        :returns: True if object is denylisted
        """
        key = self._denylist_key(obj)
        return self._denylist.get(key, 0) >= _ObjectList.BLACKLIST_COUNT

    def remove(self, obj):
        """
        Remove an object from the list.

        :param obj: vmmLibvirtObject to remove
        :returns: True if object removed, False if object was not found
        """
        with self._lock:
            # Identity check is sufficient here, since we should never be
            # asked to remove an object that wasn't at one point in the list.
            if obj not in self._objects:
                return self.remove_denylist(obj)

            self._objects.remove(obj)
            return True

    def add(self, obj):
        """
        Add an object to the list.

        :param obj: vmmLibvirtObject to add
        :returns: True if object added, False if object already in the list
        """
        with self._lock:
            # We don't look up based on identity here, to prevent tick()
            # races from adding the same domain twice
            #
            # We don't use lookup_object here since we need to hold the
            # lock the whole time to prevent a 'time of check' issue
            for checkobj in self._objects:
                if checkobj.__class__ == obj.__class__ and checkobj.get_name() == obj.get_name():
                    return False

            self._objects.append(obj)
            return True

    def get_objects_for_class(self, classobj):
        """
        Return all objects over the passed vmmLibvirtObject class
        """
        with self._lock:
            return [o for o in self._objects if o.__class__ is classobj]

    def lookup_object(self, classobj, name):
        """
        Lookup an object with the passed classobj + name
        """
        # Doesn't require locking, since get_objects_for_class covers us
        for obj in self.get_objects_for_class(classobj):
            if obj.get_name() == name:
                return obj
        return None

    def all_objects(self):
        with self._lock:
            return self._objects[:]


class vmmConnection(vmmGObject):
    __gsignals__ = {
        "vm-added": (vmmGObject.RUN_FIRST, None, [object]),
        "vm-removed": (vmmGObject.RUN_FIRST, None, [object]),
        "net-added": (vmmGObject.RUN_FIRST, None, [object]),
        "net-removed": (vmmGObject.RUN_FIRST, None, [object]),
        "pool-added": (vmmGObject.RUN_FIRST, None, [object]),
        "pool-removed": (vmmGObject.RUN_FIRST, None, [object]),
        "nodedev-added": (vmmGObject.RUN_FIRST, None, [object]),
        "nodedev-removed": (vmmGObject.RUN_FIRST, None, [object]),
        "resources-sampled": (vmmGObject.RUN_FIRST, None, []),
        "state-changed": (vmmGObject.RUN_FIRST, None, []),
        "open-completed": (vmmGObject.RUN_FIRST, None, [object]),
    }

    (_STATE_DISCONNECTED, _STATE_CONNECTING, _STATE_ACTIVE) = range(1, 4)

    def __init__(self, uri):
        self._uri = uri
        vmmGObject.__init__(self)

        self._state = self._STATE_DISCONNECTED
        self._backend = virtinst.VirtinstConnection(self._uri)
        self._closing = False

        # Error strings are stored here if open() fails
        self.connect_error = None

        self._init_object_count = None
        self._init_object_event = None

        self.using_domain_events = False
        self._domain_cb_ids = []
        self.using_network_events = False
        self._network_cb_ids = []
        self.using_storage_pool_events = False
        self._storage_pool_cb_ids = []
        self.using_node_device_events = False
        self._node_device_cb_ids = []

        self._xml_flags = {}

        self._objects = _ObjectList()
        self.statsmanager = vmmStatsManager()

        self._stats = []
        self._hostinfo = None

        self.add_gsettings_handle(
            self._on_config_pretty_name_changed(self._config_pretty_name_changed_cb)
        )

        self._init_virtconn()

    @staticmethod
    def pretty_hv(gtype, domtype):
        """
        Convert XML <domain type='foo'> and <os><type>bar</type>
        into a more human relevant string.
        """

        gtype = gtype.lower()
        domtype = domtype.lower()

        label = domtype
        if domtype == "xen":
            if gtype == "xen":
                label = "xen (paravirt)"
            elif gtype == "hvm":
                label = "xen (fullvirt)"
        elif domtype == "test":
            if gtype == "xen":
                label = "test (xen)"
            elif gtype == "hvm":
                label = "test (hvm)"
        elif domtype == "qemu":
            label = "QEMU TCG"
        elif domtype == "kvm":
            label = "KVM"
        elif domtype == "hvf":
            label = "Hypervisor.framework"  # pragma: no cover

        return label

    def __repr__(self):
        # pylint: disable=arguments-differ
        return "<%s uri=%s id=%s>" % (self.__class__.__name__, self.get_uri(), hex(id(self)))

    #################
    # Init routines #
    #################

    def _wait_for_condition(self, compare_cb, timeout=3):
        """
        Wait for this object to emit the specified signal. Will not
        block the mainloop.
        """
        from gi.repository import Gtk

        is_main_thread = threading.current_thread().name == "MainThread"
        start_time = time.time()

        while True:
            cur_time = time.time()
            if compare_cb():
                return
            if (cur_time - start_time) >= timeout:
                return  # pragma: no cover

            if is_main_thread:
                if Gtk.events_pending():
                    Gtk.main_iteration_do(False)
                    continue

            time.sleep(0.1)

    def _init_virtconn(self):
        self._backend.cb_fetch_all_domains = lambda: [
            obj.get_xmlobj(refresh_if_nec=False) for obj in self.list_vms()
        ]
        self._backend.cb_fetch_all_pools = lambda: [
            obj.get_xmlobj(refresh_if_nec=False) for obj in self.list_pools()
        ]
        self._backend.cb_fetch_all_nodedevs = lambda: [
            obj.get_xmlobj(refresh_if_nec=False) for obj in self.list_nodedevs()
        ]

        def fetch_all_vols():
            ret = []
            for pool in self.list_pools():
                for vol in pool.get_volumes():
                    try:
                        ret.append(vol.get_xmlobj(refresh_if_nec=False))
                    except Exception as e:  # pragma: no cover
                        log.debug("Fetching volume XML failed: %s", e)
            return ret

        self._backend.cb_fetch_all_vols = fetch_all_vols

        def cache_new_pool(obj):
            if not self.is_active():
                return
            name = obj.name()
            self.schedule_priority_tick(pollpool=True)

            def compare_cb():
                return bool(self.get_pool_by_name(name))

            self._wait_for_condition(compare_cb)

        self._backend.cb_cache_new_pool = cache_new_pool

    ########################
    # General data getters #
    ########################

    def get_uri(self):
        return self._uri

    def get_backend(self):
        return self._backend

    def invalidate_caps(self):
        return self._backend.invalidate_caps()

    caps = property(lambda self: getattr(self, "_backend").caps)

    def host_memory_size(self):
        if not self._backend.is_open() or self._hostinfo is None:
            return 0
        return self._hostinfo[1] * 1024

    def host_active_processor_count(self):
        if not self._backend.is_open() or self._hostinfo is None:
            return 0  # pragma: no cover
        return self._hostinfo[2]

    ##########################
    # URI + hostname helpers #
    ##########################

    def libvirt_gethostname(self):
        return self._backend.getHostname()

    get_uri_hostname = property(lambda s: getattr(s, "_backend").get_uri_hostname)
    get_uri_username = property(lambda s: getattr(s, "_backend").get_uri_username)
    get_uri_transport = property(lambda s: getattr(s, "_backend").get_uri_transport)
    get_uri_port = property(lambda s: getattr(s, "_backend").get_uri_port)
    get_driver = property(lambda s: getattr(s, "_backend").get_uri_driver)
    is_container_only = property(lambda s: getattr(s, "_backend").is_container_only)
    is_lxc = property(lambda s: getattr(s, "_backend").is_lxc)
    is_vz = property(lambda s: getattr(s, "_backend").is_vz)
    is_xen = property(lambda s: getattr(s, "_backend").is_xen)
    is_remote = property(lambda s: getattr(s, "_backend").is_remote)
    is_qemu = property(lambda s: getattr(s, "_backend").is_qemu)
    is_qemu_privileged = property(lambda s: getattr(s, "_backend").is_qemu_privileged)
    is_qemu_unprivileged = property(lambda s: getattr(s, "_backend").is_qemu_unprivileged)
    is_test = property(lambda s: getattr(s, "_backend").is_test)
    is_unprivileged = property(lambda s: getattr(s, "_backend").is_unprivileged)

    def get_cache_dir(self):
        uri = self.get_uri().replace("/", "_")
        ret = os.path.join(self._backend.get_app_cache_dir(), uri)
        if not os.path.exists(ret):
            os.makedirs(ret, 0o755)  # pragma: no cover
        return ret

    def get_default_storage_format(self):
        raw = self.config.get_default_storage_format(raw=True)
        if raw != "default":
            return raw  # pragma: no cover

        fmt = self.config.get_default_storage_format()
        if fmt != "qcow2":
            return fmt  # pragma: no cover

        if self.support.conn_default_qcow2():
            return fmt
        return None  # pragma: no cover

    ####################################
    # Connection pretty print routines #
    ####################################

    def get_pretty_desc(self):
        """
        Return a pretty label for use in the manager view, and various
        connection lists.
        """
        if self._get_config_pretty_name():
            return self._get_config_pretty_name()

        pretty_map = {
            "esx": "ESX",
            "lxc": "LXC",
            "openvz": "OpenVZ",
            "qemu": "QEMU/KVM",
            "vbox": "Virtualbox",
            "vmware": "VMWare",
            "xen": "Xen",
        }

        hv = pretty_map.get(self.get_driver(), self.get_driver())
        hostname = self.get_uri_hostname()
        path = self.get_backend().get_uri_path()

        ret = hv

        if path == "/session":
            ret += " " + _("User session")
        elif path and path != "/system" and os.path.basename(path):
            # Used by test URIs to report what XML file they are using
            ret += " %s" % os.path.basename(path)

        if hostname:
            ret += ": %s" % hostname

        return ret

    #######################
    # API support helpers #
    #######################

    @property
    def support(self):
        return self._backend.support

    def _get_flags_helper(self, obj, key, check_func):
        ignore = obj
        flags_dict = self._xml_flags.get(key)

        if flags_dict is None:
            # Flags already set
            inact, act = check_func()
            flags_dict = {}
            flags_dict["active"] = act
            flags_dict["inactive"] = inact

            self._xml_flags[key] = flags_dict

        active_flags = flags_dict["active"]
        inactive_flags = flags_dict["inactive"]

        return (inactive_flags, active_flags)

    def get_dom_flags(self, vm):
        key = "domain"

        def check_func():
            act = 0
            inact = 0

            if self.support.domain_xml_inactive(vm):
                inact = libvirt.VIR_DOMAIN_XML_INACTIVE
            else:  # pragma: no cover
                log.debug("Domain XML inactive flag not supported.")

            if self.support.domain_xml_secure(vm):
                inact |= libvirt.VIR_DOMAIN_XML_SECURE
                act = libvirt.VIR_DOMAIN_XML_SECURE
            else:  # pragma: no cover
                log.debug("Domain XML secure flag not supported.")

            return inact, act

        return self._get_flags_helper(vm, key, check_func)

    def get_default_pool(self):
        poolxml = virtinst.StoragePool.lookup_default_pool(self.get_backend())
        if poolxml:
            for p in self.list_pools():
                if p.get_name() == poolxml.name:
                    return p
        return None

    def get_vol_by_path(self, path):
        for pool in self.list_pools():
            for vol in pool.get_volumes():
                try:
                    if vol.get_target_path() == path:
                        return vol
                except Exception as e:  # pragma: no cover
                    # Errors can happen if the volume disappeared, bug 1092739
                    log.debug("Error looking up volume from path=%s: %s", path, e)
        return None

    ###################################
    # Connection state getter/setters #
    ###################################

    def _change_state(self, newstate):
        if self._state != newstate:
            self._state = newstate
            log.debug("conn=%s changed to state=%s", self.get_uri(), self.get_state_text())
            self.emit("state-changed")

    def is_active(self):
        return self._state == self._STATE_ACTIVE

    def is_disconnected(self):
        return self._state == self._STATE_DISCONNECTED

    def is_connecting(self):
        return self._state == self._STATE_CONNECTING

    def get_state_text(self):
        if self.is_disconnected():
            return _("Disconnected")
        elif self.is_connecting():
            return _("Connecting")
        return _("Active")

    #################################
    # Libvirt object lookup methods #
    #################################

    def get_vm_by_name(self, name):
        return self._objects.lookup_object(vmmDomain, name)

    def list_vms(self):
        return self._objects.get_objects_for_class(vmmDomain)

    def get_net_by_name(self, name):
        return self._objects.lookup_object(vmmNetwork, name)

    def list_nets(self):
        return self._objects.get_objects_for_class(vmmNetwork)

    def get_pool_by_name(self, name):
        return self._objects.lookup_object(vmmStoragePool, name)

    def list_pools(self):
        return self._objects.get_objects_for_class(vmmStoragePool)

    def get_nodedev_by_name(self, name):
        return self._objects.lookup_object(vmmNodeDevice, name)

    def list_nodedevs(self):
        return self._objects.get_objects_for_class(vmmNodeDevice)

    ############################
    # nodedev helper functions #
    ############################

    def filter_nodedevs(self, devtype):
        retdevs = []
        for dev in self.list_nodedevs():
            try:
                xmlobj = dev.get_xmlobj()
            except libvirt.libvirtError as e:  # pragma: no cover
                # Libvirt nodedev XML fetching can be busted
                # https://bugzilla.redhat.com/show_bug.cgi?id=1225771
                if e.get_error_code() != libvirt.VIR_ERR_NO_NODE_DEVICE:
                    log.debug("Error fetching nodedev XML", exc_info=True)
                continue

            if devtype and xmlobj.device_type != devtype:
                continue

            retdevs.append(dev)
        return retdevs

    ###################################
    # Libvirt object creation methods #
    ###################################

    def define_domain(self, xml):
        return self._backend.defineXML(xml)

    def define_network(self, xml):
        return self._backend.networkDefineXML(xml)

    def define_pool(self, xml):
        return self._backend.storagePoolDefineXML(xml, 0)

    def rename_object(self, obj, origxml, newxml):
        if obj.is_domain():
            define_cb = self.define_domain
        elif obj.is_pool():
            define_cb = self.define_pool
        elif obj.is_network():
            define_cb = self.define_network
        else:
            raise virtinst.xmlutil.DevError(
                "rename_object helper doesn't support object class %s" % obj.__class__
            )

        # Undefine the original object
        obj.delete(force=False)

        newobj = None
        try:
            # Redefine new domain
            newobj = define_cb(newxml)
        except Exception as renameerr:
            try:
                log.debug("Error defining new name %s XML", obj.class_name(), exc_info=True)
                newobj = define_cb(origxml)
            except Exception as fixerr:  # pragma: no cover
                log.debug("Failed to redefine original %s!", obj.class_name(), exc_info=True)
                msg = _(
                    "%(object)s rename failed. Attempting to recover also "
                    "failed.\n"
                    "\n"
                    "Original error: %(origerror)s\n"
                    "\n"
                    "Recover error: %(recovererror)s"
                ) % {
                    "object": obj.class_name(),
                    "origerror": str(renameerr),
                    "recovererror": str(fixerr),
                }
                raise RuntimeError(msg) from None
            raise
        finally:
            if newobj:
                # Reinsert handle into new obj
                obj.change_name_backend(newobj)

    #########################
    # Domain event handling #
    #########################

    # Our strategy here isn't the most efficient: since we need to keep the
    # poll helpers around for compat with old libvirt, switching to a fully
    # event driven setup is hard, so we end up doing more polling than
    # necessary on most events.

    def _domain_xml_misc_event(self, conn, domain, *args):
        # Just trigger a domain XML refresh for hotplug type events
        ignore = conn
        args = list(args)
        eventstr = args.pop(-1)

        name = domain.name()
        log.debug("domain xmlmisc event: domain=%s event=%s args=%s", name, eventstr, args)
        obj = self.get_vm_by_name(name)
        if obj:
            self.idle_add(obj.recache_from_event_loop)

    def _domain_lifecycle_event(self, conn, domain, state, reason, userdata):
        ignore = conn
        ignore = userdata

        name = domain.name()
        log.debug(
            "domain lifecycle event: domain=%s %s",
            name,
            LibvirtEnumMap.domain_lifecycle_str(state, reason),
        )

        obj = self.get_vm_by_name(name)

        if obj:
            self.idle_add(obj.recache_from_event_loop)
        else:
            self.schedule_priority_tick(pollvm=True, force=True)

    def _domain_agent_lifecycle_event(self, conn, domain, state, reason, userdata):
        ignore = conn
        ignore = userdata

        name = domain.name()
        log.debug(
            "domain agent lifecycle event: domain=%s %s",
            name,
            LibvirtEnumMap.domain_agent_lifecycle_str(state, reason),
        )

        obj = self.get_vm_by_name(name)

        # This event is triggered when deleting external snapshots and it changes
        # shutoff VM into paused and makes that VM unusable until virt-manager is
        # restarted so we need to ignore it in case VM is shutoff.
        if obj and obj.is_shutoff():  # pragma: no cover
            log.debug("received agent lifecycle event but domain is shutoff, ignoring it")
            return

        if obj:
            self.idle_add(obj.recache_from_event_loop)
        else:
            self.schedule_priority_tick(pollvm=True, force=True)  # pragma: no cover

    def _network_lifecycle_event(self, conn, network, state, reason, userdata):
        ignore = conn
        ignore = userdata

        name = network.name()
        log.debug(
            "network lifecycle event: network=%s %s",
            name,
            LibvirtEnumMap.network_lifecycle_str(state, reason),
        )
        obj = self.get_net_by_name(name)

        if obj:
            self.idle_add(obj.recache_from_event_loop)
        else:
            self.schedule_priority_tick(pollnet=True, force=True)

    def _storage_pool_lifecycle_event(self, conn, pool, state, reason, userdata):
        ignore = conn
        ignore = userdata

        name = pool.name()
        log.debug(
            "storage pool lifecycle event: pool=%s %s",
            name,
            LibvirtEnumMap.storage_lifecycle_str(state, reason),
        )

        obj = self.get_pool_by_name(name)

        if obj:
            self.idle_add(obj.recache_from_event_loop)
        else:
            self.schedule_priority_tick(pollpool=True, force=True)

    def _storage_pool_refresh_event(self, conn, pool, userdata):
        ignore = conn
        ignore = userdata

        name = pool.name()
        log.debug("storage pool refresh event: pool=%s", name)

        obj = self.get_pool_by_name(name)

        if not obj:
            return

        self.idle_add(obj.refresh_pool_cache_from_event_loop)

    def _node_device_lifecycle_event(self, conn, dev, state, reason, userdata):
        ignore = conn
        ignore = userdata

        name = dev.name()
        log.debug(
            "node device lifecycle event: nodedev=%s %s",
            name,
            LibvirtEnumMap.nodedev_lifecycle_str(state, reason),
        )

        self.schedule_priority_tick(pollnodedev=True, force=True)

    def _node_device_update_event(self, conn, dev, userdata):
        ignore = conn
        ignore = userdata

        name = dev.name()
        log.debug("node device update event: nodedev=%s", name)

        obj = self.get_nodedev_by_name(name)

        if obj:
            self.idle_add(obj.recache_from_event_loop)

    def _add_conn_events(self):
        if not self.support.conn_working_xen_events():
            return  # pragma: no cover

        def _check_events_disabled():
            if self.config.CLITestOptions.no_events:
                raise RuntimeError("events disabled via cli")

        try:
            _check_events_disabled()

            self._domain_cb_ids.append(
                self.get_backend().domainEventRegisterAny(
                    None, libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE, self._domain_lifecycle_event, None
                )
            )
            self.using_domain_events = True
            log.debug("Using domain events")
        except Exception as e:
            self.using_domain_events = False
            log.debug("Error registering domain events: %s", e)

        def _add_domain_xml_event(eventname, eventval, cb=None):
            if not self.using_domain_events:
                return
            if not cb:
                cb = self._domain_xml_misc_event
            try:
                eventid = getattr(libvirt, eventname, eventval)
                self._domain_cb_ids.append(
                    self.get_backend().domainEventRegisterAny(None, eventid, cb, eventname)
                )

                if (
                    eventname == "VIR_DOMAIN_EVENT_ID_AGENT_LIFECYCLE"
                    and self.config.CLITestOptions.fake_agent_event
                ):
                    testmock.schedule_fake_agent_event(self, cb)
            except Exception as e:  # pragma: no cover
                log.debug("Error registering %s event: %s", eventname, e)

        _add_domain_xml_event("VIR_DOMAIN_EVENT_ID_BALLOON_CHANGE", 13)
        _add_domain_xml_event("VIR_DOMAIN_EVENT_ID_TRAY_CHANGE", 10)
        _add_domain_xml_event("VIR_DOMAIN_EVENT_ID_DEVICE_REMOVED", 15)
        _add_domain_xml_event("VIR_DOMAIN_EVENT_ID_DEVICE_ADDED", 19)
        _add_domain_xml_event(
            "VIR_DOMAIN_EVENT_ID_AGENT_LIFECYCLE", 18, self._domain_agent_lifecycle_event
        )
        _add_domain_xml_event("VIR_DOMAIN_EVENT_ID_METADATA_CHANGE", 23)

        try:
            _check_events_disabled()

            eventid = getattr(libvirt, "VIR_NETWORK_EVENT_ID_LIFECYCLE", 0)
            self._network_cb_ids.append(
                self.get_backend().networkEventRegisterAny(
                    None, eventid, self._network_lifecycle_event, None
                )
            )
            self.using_network_events = True
            log.debug("Using network events")
        except Exception as e:
            self.using_network_events = False
            log.debug("Error registering network events: %s", e)

        try:
            _check_events_disabled()

            eventid = getattr(libvirt, "VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE", 0)
            refreshid = getattr(libvirt, "VIR_STORAGE_POOL_EVENT_ID_REFRESH", 1)
            self._storage_pool_cb_ids.append(
                self.get_backend().storagePoolEventRegisterAny(
                    None, eventid, self._storage_pool_lifecycle_event, None
                )
            )
            self._storage_pool_cb_ids.append(
                self.get_backend().storagePoolEventRegisterAny(
                    None, refreshid, self._storage_pool_refresh_event, None
                )
            )
            self.using_storage_pool_events = True
            log.debug("Using storage pool events")
        except Exception as e:
            self.using_storage_pool_events = False
            log.debug("Error registering storage pool events: %s", e)

        try:
            _check_events_disabled()

            eventid = getattr(libvirt, "VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE", 0)
            updateid = getattr(libvirt, "VIR_NODE_DEVICE_EVENT_ID_UPDATE", 1)
            lifecycle_cb = self._node_device_lifecycle_event
            update_cb = self._node_device_update_event

            self._node_device_cb_ids.append(
                self.get_backend().nodeDeviceEventRegisterAny(None, eventid, lifecycle_cb, None)
            )
            self._node_device_cb_ids.append(
                self.get_backend().nodeDeviceEventRegisterAny(None, updateid, update_cb, None)
            )

            if self.config.CLITestOptions.fake_nodedev_event:
                testmock.schedule_fake_nodedev_event(self, lifecycle_cb, update_cb)

            self.using_node_device_events = True
            log.debug("Using node device events")
        except Exception as e:
            self.using_network_events = False
            log.debug("Error registering node device events: %s", e)

    ######################################
    # Connection closing/opening methods #
    ######################################

    def _schedule_close(self):
        self._closing = True
        self.idle_add(self.close)

    def close(self):
        if not self.is_disconnected():
            log.debug("conn.close() uri=%s", self.get_uri())
        self._closing = True

        try:
            if self._backend.is_open():
                for eid in self._domain_cb_ids:
                    self._backend.domainEventDeregisterAny(eid)
                for eid in self._network_cb_ids:
                    self._backend.networkEventDeregisterAny(eid)
                for eid in self._storage_pool_cb_ids:
                    self._backend.storagePoolEventDeregisterAny(eid)
                for eid in self._node_device_cb_ids:
                    self._backend.nodeDeviceEventDeregisterAny(eid)
        except Exception:  # pragma: no cover
            log.debug("Failed to deregister events in conn cleanup", exc_info=True)
        finally:
            self._domain_cb_ids = []
            self._network_cb_ids = []
            self._storage_pool_cb_ids = []
            self._node_device_cb_ids = []

        self._stats = []

        if self._init_object_event:
            self._init_object_event.clear()  # pragma: no cover

        for obj in self._objects.all_objects():
            self._objects.remove(obj)
            try:
                self._remove_object_signal(obj)
                obj.cleanup()
            except Exception as e:  # pragma: no cover
                log.debug("Failed to cleanup %s: %s", obj, e)
        self._objects.cleanup()
        self._objects = _ObjectList()

        closeret = self._backend.close()
        if closeret == 1:
            log.debug(  # pragma: no cover
                "LEAK: conn close() returned 1, meaning refs may have leaked."
            )

        self._change_state(self._STATE_DISCONNECTED)
        self._closing = False

    def _cleanup(self):
        self.close()

        self._objects = None
        self._backend.cb_fetch_all_domains = None
        self._backend.cb_fetch_all_pools = None
        self._backend.cb_fetch_all_nodedevs = None
        self._backend.cb_fetch_all_vols = None
        self._backend.cb_cache_new_pool = None

        self.statsmanager.cleanup()
        self.statsmanager = None

    def open(self):
        if not self.is_disconnected():
            return  # pragma: no cover

        self._change_state(self._STATE_CONNECTING)

        log.debug("Scheduling background open thread for %s", self.get_uri())
        self._start_thread(self._open_thread, "Connect %s" % self.get_uri())

    def _do_open(self):
        warnconsole = False
        libvirt_error_code = None
        libvirt_error_message = None
        exc = None

        try:
            cb = connectauth.creds_dialog
            data = self
            if self.config.CLITestOptions.fake_openauth:
                testmock.fake_openauth(self, cb, data)
            if self.config.CLITestOptions.fake_session_error:
                lerr = libvirt.libvirtError("fake session error")
                lerr.err = [libvirt.VIR_ERR_AUTH_FAILED, None, "fake session error not authorized"]
                raise lerr
            self._backend.open(cb, data)
            return True, None
        except Exception as e:
            exc = e
            tb = "".join(traceback.format_exc())
            if isinstance(exc, libvirt.libvirtError):
                # pylint: disable=no-member
                libvirt_error_code = exc.get_error_code()
                libvirt_error_message = exc.get_error_message()

        if libvirt_error_code == getattr(
            libvirt, "VIR_ERR_AUTH_CANCELLED", None
        ):  # pragma: no cover
            log.debug("User cancelled auth, not raising any error.")
            return False, None

        if (
            libvirt_error_code == libvirt.VIR_ERR_AUTH_FAILED
            and "not authorized" in libvirt_error_message.lower()
        ):
            log.debug(
                "Looks like we might have failed policykit "
                "auth. Checking to see if we have a valid "
                "console session"
            )
            if not self.is_remote():
                warnconsole = bool(not connectauth.do_we_have_session())
            if self.config.CLITestOptions.fake_session_error:
                warnconsole = True

        ConnectError = connectauth.connect_error(self, str(exc), tb, warnconsole)
        return False, ConnectError

    def _populate_initial_state(self):
        if not self.support.conn_domain():  # pragma: no cover
            raise RuntimeError("Connection does not support required domain listing APIs")

        if not self.support.conn_storage():  # pragma: no cover
            log.debug("Connection doesn't seem to support storage APIs.")
        if not self.support.conn_network():  # pragma: no cover
            log.debug("Connection doesn't seem to support network APIs.")
        if not self.support.conn_nodedev():  # pragma: no cover
            log.debug("Connection doesn't seem to support nodedev APIs.")

        self._add_conn_events()

        try:
            self._backend.setKeepAlive(20, 1)
        except Exception as e:
            log.debug("Failed to setKeepAlive: %s", str(e))

        # The initial tick will set up a threading event that will only
        # trigger after all the polled libvirt objects are fully initialized.
        # That way we only report the connection is open when everything is
        # nicely setup for the rest of the app.

        self._init_object_event = threading.Event()
        self._init_object_count = 0

        self.schedule_priority_tick(
            stats_update=True,
            pollvm=True,
            pollnet=True,
            pollpool=True,
            pollnodedev=True,
            force=True,
            initial_poll=True,
        )

        self._init_object_event.wait()
        self._init_object_event = None
        self._init_object_count = None

        # Try to create the default storage pool
        # We need this after events setup so we can determine if the default
        # pool already exists
        try:
            virtinst.StoragePool.build_default_pool(self.get_backend())
        except Exception as e:  # pragma: no cover
            log.debug("Building default pool failed: %s", str(e))

    def _open_thread(self):
        try:
            is_active, ConnectError = self._do_open()
            if is_active:
                self._populate_initial_state()
        except Exception as e:  # pragma: no cover
            is_active = False
            ConnectError = connectauth.connect_error(
                self, str(e), "".join(traceback.format_exc()), False
            )

        if is_active:
            self.idle_add(self._change_state, self._STATE_ACTIVE)
        else:
            self._schedule_close()

        self.idle_emit("open-completed", ConnectError)

    #######################
    # Tick/Update methods #
    #######################

    def _remove_object_signal(self, obj):
        if obj.is_domain():
            self.emit("vm-removed", obj)
        elif obj.is_network():
            self.emit("net-removed", obj)
        elif obj.is_pool():
            self.emit("pool-removed", obj)
        elif obj.is_nodedev():
            self.emit("nodedev-removed", obj)

    def _gone_object_signals(self, gone_objects):
        """
        Responsible for signaling the UI for any updates. All possible UI
        updates need to go here to enable threading that doesn't block the
        app with long tick operations.
        """
        if not self._backend.is_open():
            return  # pragma: no cover

        for obj in gone_objects:
            class_name = obj.class_name()
            name = obj.get_name()

            if not self._objects.remove(obj):
                log.debug(
                    "Requested removal of %s=%s, but it's not in our object list.",
                    class_name,
                    name,
                )
                continue

            log.debug("%s=%s removed", class_name, name)
            self._remove_object_signal(obj)
            obj.cleanup()

    def _new_object_cb(self, obj, initialize_failed):
        if not self._backend.is_open():
            return  # pragma: no cover

        try:
            class_name = obj.class_name()

            if initialize_failed:
                log.debug("Blacklisting %s=%s", class_name, obj.get_name())
                count = self._objects.add_denylist(obj)
                log.debug("Object added in denylist, count=%d", count)
                return

            self._objects.remove_denylist(obj)
            if not self._objects.add(obj):
                log.debug(
                    "New %s=%s requested, but it's already tracked.", class_name, obj.get_name()
                )
                obj.cleanup()
                return

            if not obj.is_nodedev():
                # Skip nodedev logging since it's noisy and not interesting
                log.debug("%s=%s status=%s added", class_name, obj.get_name(), obj.run_status())
            if obj.is_domain():
                self.emit("vm-added", obj)
            elif obj.is_network():
                self.emit("net-added", obj)
            elif obj.is_pool():
                self.emit("pool-added", obj)
            elif obj.is_nodedev():
                self.emit("nodedev-added", obj)
        finally:
            if self._init_object_event:
                self._init_object_count -= 1
                if self._init_object_count <= 0:
                    self._init_object_event.set()

    def _poll(self, initial_poll, pollvm, pollnet, pollpool, pollnodedev):
        """
        Helper called from tick() to do necessary polling and return
        the relevant object lists
        """
        gone_objects = []
        preexisting_objects = []

        def _process_objects(ptype):
            if ptype == "nets":
                dopoll = pollnet
                objs = self.list_nets()
                cls = vmmNetwork
                pollcb = pollhelpers.fetch_nets
            elif ptype == "pools":
                dopoll = pollpool
                objs = self.list_pools()
                cls = vmmStoragePool
                pollcb = pollhelpers.fetch_pools
            elif ptype == "nodedevs":
                dopoll = pollnodedev
                objs = self.list_nodedevs()
                cls = vmmNodeDevice
                pollcb = pollhelpers.fetch_nodedevs
            else:
                dopoll = pollvm
                objs = self.list_vms()
                cls = vmmDomain
                pollcb = pollhelpers.fetch_vms

            keymap = dict((o.get_name(), o) for o in objs)

            def cb(obj, name):
                return cls(self, obj, name)

            if dopoll:
                gone, new, master = pollcb(self._backend, keymap, cb)
            else:
                gone, new, master = [], [], list(keymap.values())

            if initial_poll:
                self._init_object_count += len(new)

            gone_objects.extend(gone)
            preexisting_objects.extend([o for o in master if o not in new])
            new = [n for n in new if not self._objects.in_denylist(n)]
            return new

        new_vms = _process_objects("vms")
        new_nets = _process_objects("nets")
        new_pools = _process_objects("pools")
        new_nodedevs = _process_objects("nodedevs")

        # Kick off one thread per object type to handle the initial
        # XML fetching. Going any more fine grained then this probably
        # won't be that useful due to libvirt's locking structure.
        #
        # Would prefer to start refreshing some objects before all polling
        # is complete, but we need init_object_count to be fully accurate
        # before we start initializing objects

        if initial_poll and self._init_object_count == 0:
            # If the connection doesn't have any objects, new_object_cb
            # is never called and the event is never set, so let's do it here
            self._init_object_event.set()

        for newlist in [new_vms, new_nets, new_pools, new_nodedevs]:
            if not newlist:
                continue

            def cb(lst):
                for obj in lst:
                    obj.connect_once("initialized", self._new_object_cb)
                    obj.init_libvirt_state()

            self._start_thread(
                cb, "refreshing xml for new %s" % newlist[0].class_name(), args=(newlist,)
            )

        return gone_objects, preexisting_objects

    def _tick(
        self,
        stats_update=False,
        pollvm=False,
        pollnet=False,
        pollpool=False,
        pollnodedev=False,
        force=False,
        initial_poll=False,
    ):
        """
        main update function: polls for new objects, updates stats, ...

        :param force: Perform the requested polling even if async events
            are in use.
        """
        if self._closing:
            return  # pragma: no cover
        if self.is_disconnected():
            return  # pragma: no cover
        if self.is_connecting() and not force:
            return  # pragma: no cover

        # We need to set this before the event check, since stats polling
        # is independent of events
        if not pollvm:
            stats_update = False

        if self.using_domain_events and not force:
            pollvm = False
        if self.using_network_events and not force:
            pollnet = False
        if self.using_storage_pool_events and not force:
            pollpool = False
        if self.using_node_device_events and not force:
            pollnodedev = False

        self._hostinfo = self._backend.getInfo()
        if stats_update:
            self.statsmanager.cache_all_stats(self)

        gone_objects, preexisting_objects = self._poll(
            initial_poll, pollvm, pollnet, pollpool, pollnodedev
        )
        self.idle_add(self._gone_object_signals, gone_objects)

        # Only tick() pre-existing objects, since new objects will be
        # initialized asynchronously and tick() would be redundant
        for obj in preexisting_objects:
            try:
                if obj.reports_stats() and stats_update:
                    pass
                elif obj.is_domain() and not pollvm:
                    continue
                elif obj.is_network() and not pollnet:
                    continue
                elif obj.is_pool() and not pollpool:
                    continue
                elif obj.is_nodedev() and not pollnodedev:
                    continue

                if self.config.CLITestOptions.conn_crash:
                    self._backend.close()
                    e = libvirt.libvirtError("fake error")
                    e.err = [libvirt.VIR_ERR_SYSTEM_ERROR]
                    raise e

                obj.tick(stats_update=stats_update)
            except Exception as e:
                log.exception("Tick for %s failed", obj)
                if isinstance(e, libvirt.libvirtError) and (
                    getattr(e, "get_error_code")() == libvirt.VIR_ERR_SYSTEM_ERROR
                ):
                    # Try a simple getInfo call to see if conn was dropped
                    self._backend.getInfo()
                    log.debug(  # pragma: no cover
                        "vm tick raised system error but "
                        "connection doesn't seem to have dropped. "
                        "Ignoring."
                    )

        if stats_update:
            self._recalculate_stats([o for o in preexisting_objects if o.reports_stats()])
            self.idle_emit("resources-sampled")

    def _recalculate_stats(self, vms):
        if not self._backend.is_open():
            return  # pragma: no cover

        now = time.time()
        expected = self.config.get_stats_history_length()
        current = len(self._stats)
        if current > expected:
            del self._stats[expected:current]  # pragma: no cover

        mem = 0
        cpuTime = 0
        rdRate = 0
        wrRate = 0
        rxRate = 0
        txRate = 0
        diskMaxRate = self.disk_io_max_rate() or 10.0
        netMaxRate = self.network_traffic_max_rate() or 10.0

        for vm in vms:
            if not vm.is_active():
                continue

            cpuTime += vm.cpu_time()
            mem += vm.stats_memory()
            rdRate += vm.disk_read_rate()
            wrRate += vm.disk_write_rate()
            rxRate += vm.network_rx_rate()
            txRate += vm.network_tx_rate()

            netMaxRate = max(netMaxRate, vm.network_traffic_max_rate())
            diskMaxRate = max(diskMaxRate, vm.disk_io_max_rate())

        pcentHostCpu = 0
        pcentMem = mem * 100.0 / self.host_memory_size()

        if len(self._stats) > 0:
            prevTimestamp = self._stats[0]["timestamp"]
            host_cpus = self.host_active_processor_count()

            pcentHostCpu = (
                (cpuTime) * 100.0 / ((now - prevTimestamp) * 1000.0 * 1000.0 * 1000.0 * host_cpus)
            )

        pcentHostCpu = max(0.0, min(100.0, pcentHostCpu))
        pcentMem = max(0.0, min(100.0, pcentMem))

        newStats = {
            "timestamp": now,
            "memory": mem,
            "memoryPercent": pcentMem,
            "cpuTime": cpuTime,
            "cpuHostPercent": pcentHostCpu,
            "diskRdRate": rdRate,
            "diskWrRate": wrRate,
            "netRxRate": rxRate,
            "netTxRate": txRate,
            "diskMaxRate": diskMaxRate,
            "netMaxRate": netMaxRate,
        }

        self._stats.insert(0, newStats)

    def schedule_priority_tick(self, **kwargs):
        from .engine import vmmEngine

        vmmEngine.get_instance().schedule_priority_tick(self, kwargs)

    def tick_from_engine(self, *args, **kwargs):
        try:
            self._tick(*args, **kwargs)
        except Exception:
            self._schedule_close()
            raise

    ########################
    # Stats getter methods #
    ########################

    def _get_record_helper(self, record_name):
        if len(self._stats) == 0:
            return 0
        return self._stats[0][record_name]

    def _vector_helper(self, record_name, limit, ceil=100.0):
        vector = []
        statslen = self.config.get_stats_history_length() + 1
        if limit is not None:
            statslen = min(statslen, limit)  # pragma: no cover

        for i in range(statslen):
            if i < len(self._stats):
                vector.append(self._stats[i][record_name] / ceil)
            else:
                vector.append(0)

        return vector

    def stats_memory_vector(self, limit=None):
        return self._vector_helper("memoryPercent", limit)

    def host_cpu_time_vector(self, limit=None):
        return self._vector_helper("cpuHostPercent", limit)

    def stats_memory(self):
        return self._get_record_helper("memory")

    def host_cpu_time_percentage(self):
        return self._get_record_helper("cpuHostPercent")

    def guest_cpu_time_percentage(self):
        return self.host_cpu_time_percentage()

    def network_traffic_rate(self):
        return self._get_record_helper("netRxRate") + self._get_record_helper("netTxRate")

    def disk_io_rate(self):
        return self._get_record_helper("diskRdRate") + self._get_record_helper("diskWrRate")

    def network_traffic_max_rate(self):
        return self._get_record_helper("netMaxRate")

    def disk_io_max_rate(self):
        return self._get_record_helper("diskMaxRate")

    ###########################
    # Per-conn config helpers #
    ###########################

    def get_autoconnect(self):
        return self.config.get_conn_autoconnect(self.get_uri())

    def set_autoconnect(self, val):
        self.config.set_conn_autoconnect(self.get_uri(), val)

    def set_config_pretty_name(self, value):
        cfgname = self._get_config_pretty_name()
        if value == cfgname:
            return  # pragma: no cover
        if not cfgname and value == self.get_pretty_desc():
            # Don't encode the default connection value into gconf right
            # away, require the user to edit it first
            return
        self.config.set_perconn(self.get_uri(), "/pretty-name", value)

    def _get_config_pretty_name(self):
        return self.config.get_perconn(self.get_uri(), "/pretty-name")

    def _on_config_pretty_name_changed(self, *args, **kwargs):
        return self.config.listen_perconn(self.get_uri(), "/pretty-name", *args, **kwargs)

    def _config_pretty_name_changed_cb(self):
        self.emit("state-changed")

    def set_details_window_size(self, w, h):
        self.config.set_perconn(self.get_uri(), "/window-size", (w, h))

    def get_details_window_size(self):
        ret = self.config.get_perconn(self.get_uri(), "/window-size")
        return ret



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import tests
from . import lib


#############################################
# UI tests for details storage media change #
#############################################


def testMediaChange(app):
    vmname = "test-many-devices"
    app.uri = tests.utils.URIs.test_remote
    app.open(show_console=vmname)
    win = app.find_details_window(vmname, click_details=True, shutdown=True)
    hw = win.find("hw-list")
    tab = win.find("disk-tab")
    combo = win.find("media-combo")
    entry = win.find("media-entry")
    appl = win.find("config-apply")

    # Floppy + physical
    hw.find("Floppy 1", "table cell").click()
    combo.click_combo_entry()
    combo.find(r"Floppy_install_label \(/dev/fdb\)")
    lib.utils.check(lambda: entry.text == "No media detected (/dev/fda)")
    entry.click()
    entry.click_secondary_icon()
    lib.utils.check(lambda: not entry.text)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: not entry.text)

    # Enter /dev/fdb, after apply it should change to pretty label
    entry.set_text("/dev/fdb")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: entry.text == "Floppy_install_label (/dev/fdb)")

    # Specify manual path
    path = "/pool-dir/UPPER"
    entry.set_text(path)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: entry.text == path)

    # Go to Floppy 2, make sure previous path is in recent list
    hw.find("Floppy 2", "table cell").click()
    combo.click_combo_entry()
    combo.find(path)
    entry.click()
    # Use the storage browser to select new floppy storage
    tab.find("Browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "iso-vol")
    appl.click()

    # Browse for image
    hw.find("IDE CDROM 1", "table cell").click()
    combo.click_combo_entry()
    combo.find(r"Fedora12_media \(/dev/sr0\)")
    entry.click()
    tab.find("Browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "backingl1.img")
    # Check 'already in use' dialog
    appl.click()
    app.click_alert_button("already in use by", "No")
    lib.utils.check(lambda: appl.sensitive)
    appl.click()
    app.click_alert_button("already in use by", "Yes")
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: "backing" in entry.text)
    entry.set_text("")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: not entry.text)


def testMediaHotplug(app):
    """
    Test in the case of a running VM
    """
    vmname = "test-many-devices"
    app.open(show_console=vmname, extra_opts=["--test-options=test-update-device-fail"])
    win = app.find_details_window(vmname, click_details=True)
    hw = win.find("hw-list")
    entry = win.find("media-entry")
    appl = win.find("config-apply")

    hw.find("IDE CDROM 1", "table cell").click()
    lib.utils.check(lambda: not entry.text)
    # Catch path does not exist error
    entry.set_text("/dev/sr7")
    appl.click()
    app.click_alert_button("non-existent path '/dev/sr7", "Close")

    # Check relative path while we are at it
    path = "virt-install"
    entry.set_text(path)
    appl.click()
    app.click_alert_button("changes will take effect", "OK")
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: not entry.text)

    # Shutdown the VM, verify change shows up
    win.find("Shut Down", "push button").click()
    run = win.find("Run", "push button")
    lib.utils.check(lambda: run.sensitive)
    lib.utils.check(lambda: entry.text == os.path.abspath(path))



#!/usr/bin/env python3

import os
import shutil
import sys

file_name = sys.argv[1]

meson_build_root = os.environ["MESON_BUILD_ROOT"]
meson_dist_root = os.environ["MESON_DIST_ROOT"]

shutil.copy(os.path.join(meson_build_root, file_name), os.path.join(meson_dist_root, file_name))



#
# XML API wrappers
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import libxml2

from . import xmlutil
from .logger import log

# pylint: disable=protected-access


class _XPathSegment:
    """
    Class representing a single 'segment' of an xpath string. For example,
    the xpath:

        ./qemu:foo/bar[1]/baz[@somepro='someval']/@finalprop

    will be split into the following segments:

        #1: nodename=., fullsegment=.
        #2: nodename=foo, nsname=qemu, fullsegment=qemu:foo
        #3: nodename=bar, condition_num=1, fullsegment=bar[1]
        #4: nodename=baz, condition_prop=somepro, condition_val=someval,
                fullsegment=baz[@somepro='somval']
        #5: nodename=finalprop, is_prop=True, fullsegment=@finalprop
    """

    def __init__(self, fullsegment):
        self.fullsegment = fullsegment
        self.nodename = fullsegment

        self.condition_prop = None
        self.condition_val = None
        self.condition_num = None
        if "[" in self.nodename:
            self.nodename, cond = self.nodename.strip("]").split("[")
            if "=" in cond:
                (cprop, cval) = cond.split("=")
                self.condition_prop = cprop.strip("@")
                self.condition_val = cval.strip("'")
            elif cond.isdigit():
                self.condition_num = int(cond)

        self.is_prop = self.nodename.startswith("@")
        if self.is_prop:
            self.nodename = self.nodename[1:]

        self.nsname = None
        if ":" in self.nodename:
            self.nsname, self.nodename = self.nodename.split(":")


class _XPath:
    """
    Helper class for performing manipulations of XPath strings. Splits
    the xpath into segments.
    """

    def __init__(self, fullxpath):
        self.fullxpath = fullxpath
        self.segments = []
        for s in self.fullxpath.split("/"):
            if s == "..":
                # Resolve and flatten .. in xpaths
                self.segments = self.segments[:-1]
                continue
            self.segments.append(_XPathSegment(s))

        self.is_prop = self.segments[-1].is_prop
        self.propname = self.is_prop and self.segments[-1].nodename or None
        if self.is_prop:
            self.segments = self.segments[:-1]
        self.xpath = self.join(self.segments)

    @staticmethod
    def join(segments):
        return "/".join(s.fullsegment for s in segments)

    def parent_xpath(self):
        return self.join(self.segments[:-1])


class _XMLBase:
    NAMESPACES = {}

    @classmethod
    def register_namespace(cls, nsname, uri):
        cls.NAMESPACES[nsname] = uri

    def copy_api(self):
        raise NotImplementedError()

    def count(self, xpath):
        raise NotImplementedError()

    def _find(self, fullxpath):
        raise NotImplementedError()

    def _node_tostring(self, node):
        raise NotImplementedError()

    def _node_get_text(self, node):
        raise NotImplementedError()

    def _node_set_text(self, node, setval):
        raise NotImplementedError()

    def _node_get_property(self, node, propname):
        raise NotImplementedError()

    def _node_set_property(self, node, propname, setval):
        raise NotImplementedError()

    def _node_new(self, xpathseg, parentnode):
        raise NotImplementedError()

    def _node_add_child(self, parentxpath, parentnode, newnode):
        raise NotImplementedError()

    def _node_remove_child(self, parentnode, childnode):
        raise NotImplementedError()

    def _node_replace_child(self, xpath, newnode):
        raise NotImplementedError()

    def _node_from_xml(self, xml):
        raise NotImplementedError()

    def _node_has_content(self, node):
        raise NotImplementedError()

    def _node_get_name(self, node):
        raise NotImplementedError()

    def node_clear(self, xpath):
        raise NotImplementedError()

    def _sanitize_xml(self, xml):
        raise NotImplementedError()

    def get_xml(self, xpath):
        node = self._find(xpath)
        if node is None:
            return ""
        return self._sanitize_xml(self._node_tostring(node))

    def get_xpath_content(self, xpath, is_bool):
        node = self._find(xpath)
        if node is None:
            return None
        if is_bool:
            return True
        xpathobj = _XPath(xpath)
        if xpathobj.is_prop:
            return self._node_get_property(node, xpathobj.propname)
        return self._node_get_text(node)

    def set_xpath_content(self, xpath, setval):
        node = self._find(xpath)
        if setval is False:
            # Boolean False, means remove the node entirely
            self.node_force_remove(xpath)
        elif setval is None:
            if node is not None:
                self._node_set_content(xpath, node, None)
            self._node_remove_empty(xpath)
        else:
            if node is None:
                node = self._node_make_stub(xpath)

            if setval is True:
                # Boolean property, creating the node is enough
                return
            self._node_set_content(xpath, node, setval)

    def node_add_xml(self, xml, xpath):
        newnode = self._node_from_xml(xml)
        parentnode = self._node_make_stub(xpath)
        self._node_add_child(xpath, parentnode, newnode)

    def node_replace_xml(self, xpath, xml):
        """
        Replace the node at xpath with the passed in xml
        """
        newnode = self._node_from_xml(xml)
        self._node_replace_child(xpath, newnode)

    def node_force_remove(self, fullxpath):
        """
        Remove the element referenced at the passed xpath, regardless
        of whether it has children or not, and then clean up the XML
        chain
        """
        xpathobj = _XPath(fullxpath)
        parentnode = self._find(xpathobj.parent_xpath())
        childnode = self._find(fullxpath)
        if parentnode is None or childnode is None:
            return
        self._node_remove_child(parentnode, childnode)

    def validate_root_name(self, expected_root_name):
        rootname = self._node_get_name(self._find("."))
        if rootname == expected_root_name:
            return
        raise RuntimeError(
            _(
                "XML did not have expected root element name "
                "'%(expectname)s', found '%(foundname)s'"
            )
            % {"expectname": expected_root_name, "foundname": rootname}
        )

    def _node_set_content(self, xpath, node, setval):
        xpathobj = _XPath(xpath)
        if setval is not None:
            setval = str(setval)
        if xpathobj.is_prop:
            self._node_set_property(node, xpathobj.propname, setval)
        else:
            self._node_set_text(node, setval)

    def _node_make_stub(self, fullxpath):
        """
        Build all nodes for the passed xpath. For example, if XML is <foo/>,
        and xpath=./bar/@baz, after this function the XML will be:

          <foo>
            <bar baz=''/>
          </foo>

        And the node pointing to @baz will be returned, for the caller to
        do with as they please.

        There's also special handling to ensure that setting
        xpath=./bar[@baz='foo']/frob will create

          <bar baz='foo'>
            <frob></frob>
          </bar>

        Even if <bar> didn't exist before. So we fill in the dependent property
        expression values
        """
        xpathobj = _XPath(fullxpath)
        parentxpath = "."
        parentnode = self._find(parentxpath)
        if not parentnode:
            raise xmlutil.DevError("Did not find XML root node for xpath=%s" % fullxpath)

        for xpathseg in xpathobj.segments[1:]:
            oldxpath = parentxpath
            parentxpath += "/%s" % xpathseg.fullsegment
            tmpnode = self._find(parentxpath)
            if tmpnode is not None:
                # xpath node already exists, nothing to create yet
                parentnode = tmpnode
                continue

            newnode = self._node_new(xpathseg, parentnode)
            self._node_add_child(oldxpath, parentnode, newnode)
            parentnode = newnode

            # For a conditional xpath like ./foo[@bar='baz'],
            # we also want to implicitly set <foo bar='baz'/>
            if xpathseg.condition_prop:
                self._node_set_property(parentnode, xpathseg.condition_prop, xpathseg.condition_val)

        return parentnode

    def _node_remove_empty(self, fullxpath):
        """
        Walk backwards up the xpath chain, and remove each element
        if it doesn't have any children or attributes, so we don't
        leave stale elements in the XML
        """
        xpathobj = _XPath(fullxpath)
        segments = xpathobj.segments[:]
        parent = None
        while segments:
            xpath = _XPath.join(segments)
            segments.pop()
            child = parent
            parent = self._find(xpath)
            if parent is None:
                break
            if child is None:
                continue
            if self._node_has_content(child):
                break

            self._node_remove_child(parent, child)


def node_is_text(n):
    return bool(n and n.type == "text")


class _Libxml2API(_XMLBase):
    def __init__(self, xml):
        _XMLBase.__init__(self)

        # Use of gtksourceview in virt-manager changes this libxml
        # global setting which messes up whitespace after parsing.
        # We can probably get away with calling this less but it
        # would take some investigation
        libxml2.keepBlanksDefault(1)

        self._doc = libxml2.parseDoc(xml)
        self._ctx = self._doc.xpathNewContext()
        self._ctx.setContextNode(self._doc.children)
        for key, val in self.NAMESPACES.items():
            self._ctx.xpathRegisterNs(key, val)

    def __del__(self):
        if not hasattr(self, "_doc"):
            # In case we error when parsing the doc
            return
        self._doc.freeDoc()
        self._doc = None
        self._ctx.xpathFreeContext()
        self._ctx = None

    def _sanitize_xml(self, xml):
        if not xml.endswith("\n") and "\n" in xml:
            xml += "\n"
        return xml

    def copy_api(self):
        return _Libxml2API(self._doc.children.serialize())

    def _find(self, fullxpath):
        xpath = _XPath(fullxpath).xpath
        try:
            node = self._ctx.xpathEval(xpath)
        except Exception as e:
            log.debug("fullxpath=%s xpath=%s eval failed", fullxpath, xpath, exc_info=True)
            raise RuntimeError("%s %s" % (fullxpath, str(e))) from None
        return node and node[0] or None

    def count(self, xpath):
        return len(self._ctx.xpathEval(xpath))

    def _node_tostring(self, node):
        return node.serialize()

    def _node_from_xml(self, xml):
        return libxml2.parseDoc(xml).children

    def _node_get_text(self, node):
        return node.content

    def _node_set_text(self, node, setval):
        if setval is not None:
            setval = xmlutil.xml_escape(setval)
        node.setContent(setval)

    def _node_get_property(self, node, propname):
        prop = node.hasProp(propname)
        if prop:
            return prop.content

    def _node_set_property(self, node, propname, setval):
        if setval is None:
            prop = node.hasProp(propname)
            if prop:
                prop.unlinkNode()
                prop.freeNode()
        else:
            node.setProp(propname, setval)

    def _node_new(self, xpathseg, parentnode):
        newnode = libxml2.newNode(xpathseg.nodename)
        if not xpathseg.nsname:
            return newnode

        def _find_parent_ns():
            parent = parentnode
            while parent:
                for ns in xmlutil.listify(parent.nsDefs()):
                    if ns.name == xpathseg.nsname:
                        return ns
                parent = parent.get_parent()

        ns = _find_parent_ns()
        if not ns:
            ns = newnode.newNs(self.NAMESPACES[xpathseg.nsname], xpathseg.nsname)
        newnode.setNs(ns)
        return newnode

    def node_clear(self, xpath):
        node = self._find(xpath)
        if node:
            propnames = [p.name for p in (node.properties or [])]
            for p in propnames:
                node.unsetProp(p)
            node.setContent(None)

    def _node_has_content(self, node):
        return node.type == "element" and (node.children or node.properties)

    def _node_get_name(self, node):
        return node.name

    def _node_remove_child(self, parentnode, childnode):
        node = childnode

        # Look for preceding whitespace and remove it
        white = node.get_prev()
        if node_is_text(white):
            white.unlinkNode()
            white.freeNode()

        node.unlinkNode()
        node.freeNode()
        if all([node_is_text(n) for n in parentnode.children]):
            parentnode.setContent(None)

    def _node_add_child(self, parentxpath, parentnode, newnode):
        ignore = parentxpath
        if not node_is_text(parentnode.get_last()):
            prevsib = parentnode.get_prev()
            if node_is_text(prevsib):
                newlast = libxml2.newText(prevsib.content)
            else:
                newlast = libxml2.newText("\n")
            parentnode.addChild(newlast)

        endtext = parentnode.get_last().content
        parentnode.addChild(libxml2.newText("  "))
        parentnode.addChild(newnode)
        parentnode.addChild(libxml2.newText(endtext))

    def _node_replace_child(self, xpath, newnode):
        oldnode = self._find(xpath)
        oldnode.replaceNode(newnode)


XMLAPI = _Libxml2API



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import unittest

import tests.utils
from . import lib


def _stop_vm(win):
    run = win.find("Run", "push button")
    win.find("Shut Down", "push button").click()
    lib.utils.check(lambda: run.sensitive)


def _start_vm(win):
    run = win.find("Run", "push button")
    run.click()
    lib.utils.check(lambda: not run.sensitive)


def table_cell_click(app, listw, text):
    """
    Find the table cell matching 'text'. If it's not on screen,
    use ctrl+f searching to quickly select it.
    """
    c = listw.find(text, "table cell")
    if not c.onscreen:
        listw.point()
        listw.click()
        app.rawinput.keyCombo("<ctrl>f")
        searchwin = app.find_window(None, roleName="window")
        searchentry = searchwin.find(None, "text")
        searchentry.set_text(text)
        c.check_onscreen()
        lib.utils.check(lambda: c.state_selected)
        app.rawinput.pressKey("Enter")
    c.click()
    return c


def _select_hw(app, win, hwname, tabname):
    table_cell_click(app, win.find("hw-list"), hwname)
    tab = win.find(tabname, None)
    lib.utils.check(lambda: tab.showing)
    return tab


#################################################
# UI tests for virt-manager's VM details window #
#################################################


def _testSmokeTest(app, vmname):
    """
    Open the VM with all the crazy hardware and just verify that each
    HW panel shows itself without raising any error.
    """
    win = app.manager_open_details(vmname)
    lst = win.find("hw-list", "table")
    lib.utils.walkUIList(app, win, lst, lambda: False)

    # Select XML editor, and reverse walk the list
    win.find("XML", "page tab").click()
    lib.utils.walkUIList(app, win, lst, lambda: False, reverse=True)
    return win


def testDetailsHardwareSmokeTest(app):
    _testSmokeTest(app, "test-many-devices")


def testDetailsHardwareSmokeTestAlternate(app):
    app.open(keyfile="allstats.ini")
    win = _testSmokeTest(app, "test alternate devs title")
    win.find("Details", "page tab").click()
    _select_hw(app, win, "Performance", "performance-tab")
    app.sleep(1)  # wait for polling to trigger perf updates


def _testRename(app, win, origname, newname):
    # Ensure the Overview page is the first selected
    win.find("Hypervisor Details", "label")
    win.find("Overview", "table cell").click()

    oldcell = app.root.find_fuzzy(origname, "table cell")
    badname = "foo/bar"
    win.find("Name:", "text").set_text(badname)
    appl = win.find("config-apply")
    appl.click()
    app.click_alert_button(badname, "Close")

    # Actual name change
    win.find("Name:", "text").set_text(newname)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Confirm lists were updated
    app.find_window("%s on" % newname)
    app.root.find_fuzzy(newname, "table cell")

    # Make sure the old entry is gone
    if origname != newname:
        lib.utils.check(lambda: origname not in oldcell.name)


def testDetailsRenameSimple(app):
    """
    Rename a simple VM
    """
    origname = "test-clone-simple"
    win = app.manager_open_details(origname)
    _testRename(app, win, origname, origname)
    _testRename(app, win, origname, "test-new-name")


def testDetailsRenameNVRAM(app):
    """
    Rename a VM that will trigger the nvram behavior
    """
    origname = "test-many-devices"
    win = app.manager_open_details(origname, shutdown=True)
    _testRename(app, win, origname, "test-new-name")


def testDetailsStateMisc(app):
    """
    Test state changes and unapplied changes warnings
    """
    app.uri = tests.utils.URIs.kvm_x86
    win = app.manager_open_details("test", shutdown=True)
    fmenu = win.find("File", "menu")
    fmenu.click()
    fmenu.find("View Manager").click()
    # Double run to hit a show() codepath
    win = app.manager_open_details("test")
    lib.utils.check(lambda: win.active)
    appl = win.find("config-apply", "push button")

    # View Manager option
    win.find("File", "menu").click()
    win.find("View Manager", "menu item").click()
    app.topwin.window_close()
    lib.utils.check(lambda: win.active)

    # Make a change and then trigger unapplied change warning
    tab = _select_hw(app, win, "Overview", "overview-tab")
    tab.find("Name:", "text").set_text("")
    lib.utils.check(lambda: appl.sensitive)
    run = win.find("Run", "push button")
    run.click()
    # Trigger apply error to hit some code paths
    app.click_alert_button("unapplied changes", "Yes")
    app.click_alert_button("name must be specified", "Close")
    lib.utils.check(lambda: run.sensitive)
    consolebtn = win.find("Console", "radio button")
    consolebtn.click()
    app.click_alert_button("unapplied changes", "Yes")
    app.click_alert_button("name must be specified", "Close")
    lib.utils.check(lambda: not consolebtn.checked)

    # Test the pause toggle
    win.find("config-cancel").click()
    run.click()
    lib.utils.check(lambda: not run.sensitive)
    pause = win.find("Pause", "toggle button")
    pause.click()
    lib.utils.check(lambda: pause.checked)
    pause.click()
    lib.utils.check(lambda: not pause.checked)
    lib.utils.check(lambda: win.active)


def testDetailsEditDomain1(app):
    """
    Test overview, memory
    """
    app.uri = tests.utils.URIs.kvm_x86
    win = app.manager_open_details("test")
    appl = win.find("config-apply", "push button")

    # Overview description
    tab = _select_hw(app, win, "Overview", "overview-tab")
    tab.find("Description:", "text").set_text("hey new description")
    tab.find("Title:", "text").set_text("hey new title")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Memory
    tab = _select_hw(app, win, "Memory", "memory-tab")
    curmem = tab.find("Current allocation:", "spin button")
    maxmem = tab.find("Maximum allocation:", "spin button")
    curmem.set_text("2000")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    curmem.set_text("50000")
    lib.utils.check(lambda: maxmem.text == "50000")
    curmem.set_text("5000")
    lib.utils.check(lambda: maxmem.text == "50000")
    maxmem.set_text("1500")
    appl.click()
    app.click_alert_button("changes will take effect", "OK")
    lib.utils.check(lambda: not appl.sensitive)

    # There's no hotplug operations after this point
    _stop_vm(win)
    lib.utils.check(lambda: curmem.text == "1500")
    lib.utils.check(lambda: maxmem.text == "1500")

    # Change the 'Shared mem' setting, verify the XML change,
    # make sure unsetting it works too
    starting_xml = lib.utils.get_xmleditor_xml(app, win)
    tab.find("Enable shared", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    new_xml = lib.utils.get_xmleditor_xml(app, win)
    assert 'source type="memfd"' in new_xml
    tab.find("Enable shared", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    assert lib.utils.get_xmleditor_xml(app, win) == starting_xml


def testDetailsEditCPU(app):
    app.uri = tests.utils.URIs.kvm_x86_cpu_insecure
    win = app.manager_open_details("test")
    appl = win.find("config-apply", "push button")
    _stop_vm(win)

    # Static CPU config
    # more cpu config: host-passthrough, copy, clear CPU, manual
    tab = _select_hw(app, win, "CPUs", "cpu-tab")
    tab.find("cpu-model").click_combo_entry()
    tab.find_fuzzy("Clear CPU", "menu item").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab.find("cpu-model").click_combo_entry()
    tab.find("coreduo", "menu item").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab.find_fuzzy("CPU security", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab.find("cpu-model").click_combo_entry()
    tab.find("Application Default", "menu item").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    copyhost = tab.find("Copy host", "check box")
    lib.utils.check(lambda: copyhost.checked)
    copyhost.click()
    tab.find("cpu-model").click_combo_entry()
    tab.find("Hypervisor Default", "menu item").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab.find("cpu-model").find(None, "text").text = "foobar"
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab.find("cpu-model").find(None, "text").text = "host-passthrough"
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # vCPUs
    tab.find("vCPU allocation:", "spin button").set_text("50")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # CPU topology
    tab.find_fuzzy("Topology", "toggle button").click_expander()
    tab.find_fuzzy("Manually set", "check").click()
    sockets = tab.find("Sockets:", "spin button")
    sockets.typeText("8")
    tab.find("Cores:", "spin button").typeText("2")
    tab.find("Threads:", "spin button").typeText("2")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Confirm VCPUs were adjusted
    vcpualloc = tab.find_fuzzy("vCPU allocation", "spin")
    lib.utils.check(lambda: vcpualloc.text == "32")

    # Unset topology
    tab.find_fuzzy("Manually set", "check").click()
    lib.utils.check(lambda: not sockets.sensitive)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)


def testDetailsEditDomain2(app):
    """
    Test boot and OS pages
    """
    win = app.manager_open_details("test-many-devices")
    appl = win.find("config-apply", "push button")
    _stop_vm(win)

    # OS edits
    tab = _select_hw(app, win, "OS information", "os-tab")
    entry = tab.find("oslist-entry")
    lib.utils.check(lambda: entry.text == "Fedora")
    entry.click()
    app.rawinput.pressKey("Down")
    popover = win.find("oslist-popover")
    popover.find("include-eol").click()
    entry.set_text("fedora12")
    popover.find_fuzzy("fedora12").bring_on_screen().click()
    lib.utils.check(lambda: not popover.visible)
    lib.utils.check(lambda: entry.text == "Fedora 12")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    lib.utils.check(lambda: entry.text == "Fedora 12")

    # Boot tweaks
    def check_bootorder(c):
        # Click the bootlist checkbox, which is hard to find in the tree
        x = c.position[0] - 30
        y = c.position[1] + c.size[1] / 2
        button = 1
        app.rawinput.click(x, y, button)

    tab = _select_hw(app, win, "Boot Options", "boot-tab")
    tab.find_fuzzy("Start virtual machine on host", "check box").click()
    tab.find("Enable boot menu", "check box").click()
    tab.find("SCSI Disk 1", "table cell").click()
    tab.find("boot-movedown", "push button").click()
    tab.find("Floppy 1", "table cell").click()
    tab.find("boot-moveup", "push button").click()
    check_bootorder(tab.find("NIC :33:44", "table cell"))
    check_bootorder(tab.find("PCI 0003:", "table cell"))
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Kernel boot
    tab.find_fuzzy("Direct kernel boot", "toggle button").click_expander()
    tab.find_fuzzy("Enable direct kernel", "check box").click()

    tab.find("Kernel args:", "text").set_text("console=ttyS0")
    appl.click()
    app.click_alert_button("arguments without specifying", "OK")
    lib.utils.check(lambda: win.active)

    initrd = tab.find("Initrd path:", "text")
    tab.find("initrd-browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "backingl1.img")
    lib.utils.check(lambda: win.active)
    lib.utils.check(lambda: "backing" in initrd.text)
    appl.click()
    app.click_alert_button("initrd without specifying", "OK")
    lib.utils.check(lambda: win.active)

    tab.find("kernel-browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "bochs-vol")
    lib.utils.check(lambda: win.active)
    kernelpath = tab.find("Kernel path:", "text")
    lib.utils.check(lambda: "bochs" in kernelpath.text)

    dtb = tab.find("DTB path:", "text")
    tab.find("dtb-browse", "push button").click()
    app.select_storagebrowser_volume("pool-dir", "iso-vol")
    lib.utils.check(lambda: win.active)
    lib.utils.check(lambda: "iso-vol" in dtb.text)

    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Now disable kernel, but verify that we keep the values in the UI
    tab.find_fuzzy("Enable direct kernel", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab = _select_hw(app, win, "OS information", "os-tab")
    tab = _select_hw(app, win, "Boot Options", "boot-tab")
    lib.utils.check(lambda: "backing" in initrd.text)


def testDetailsAlternateEdits(app):
    """
    Some specific handling via test-alternate-devs
    """
    win = app.manager_open_details("test alternate devs title")

    # tests the console dup removal
    _select_hw(app, win, "Serial 1", "char-tab")
    win.find("config-remove").click()
    app.click_alert_button("Are you sure", "Yes")
    app.click_alert_button("take effect after", "OK")
    _stop_vm(win)


def testDetailsEmptyBoot(app):
    """
    Check boot handling when VM has no devices
    """
    win = app.manager_open_details("test-state-crashed")
    _select_hw(app, win, "Boot Options", "boot-tab")
    win.find("No bootable devices")

    # Add in switching back to the console view to hit a vmwindow path
    win.find("Console", "radio button").click()


def testDetailsEditDiskNet(app):
    """
    Test disk and network devices
    """
    win = app.manager_open_details("test-many-devices")
    appl = win.find("config-apply", "push button")

    # Quick test to hit some serialcon.py paths
    viewmenu = win.find("^View$", "menu")
    viewmenu.click()
    textmenu = viewmenu.find("Consoles", "menu")
    textmenu.point()
    conitem = textmenu.find("Serial 1")
    lib.utils.check(lambda: not conitem.sensitive)
    viewmenu.click()

    _stop_vm(win)

    # Disk options
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    tab.find("Advanced options", "toggle button").click_expander()
    tab.find("Shareable:", "check box").click()
    tab.find("Readonly:", "check box").click()
    tab.find("Serial:", "text").set_text("1234-ABCD")
    tab.combo_select("Cache mode:", "unsafe")
    tab.combo_select("Discard mode:", "unmap")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Network values w/ macvtap manual
    tab = _select_hw(app, win, "NIC :54:32:10", "network-tab")
    tab.find("IP address", "push button").click()
    src = tab.find("net-source")
    src.click()
    app.rawinput.pressKey("Home")
    tab.find_fuzzy("Macvtap device...", "menu item").bring_on_screen().click()
    tab.find("Device name:", "text").set_text("fakedev12")
    tab.combo_select("Device model:", "rtl8139")
    tab.find("Link state:", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Manual bridge
    src.click()
    tab.find_fuzzy("Bridge device...", "menu item").bring_on_screen().click()
    tab.find("Device name:", "text").set_text("")
    appl.click()
    # Check validation error
    app.click_alert_button("Error changing VM configuration", "Close")
    tab.find("Device name:", "text").set_text("zbr0")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Portgroup
    src.click()
    tab.find_fuzzy("plainbridge-portgroups", "menu item").bring_on_screen().click()
    c = tab.find_fuzzy("Portgroup:", "combo box")
    c.click_combo_entry()
    c.find("sales").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)


def testDetailsNetIPAddress(app):
    """
    Test all the IP code paths with a few mock cases
    """
    win = app.manager_open_details("test-many-devices")

    def check_ip(*args):
        for ip in args:
            tab.find_fuzzy(ip, "label")

    # First case has a virtual network, so hits the leases path
    tab = _select_hw(app, win, "NIC :54:32:10", "network-tab")
    check_ip("10.0.0.2", "fd00:beef::2")
    tab.find("IP address:", "push button").click()
    check_ip("10.0.0.2", "fd00:beef::2")

    # Next case has a missing virtual network, so hits the arp path
    tab = _select_hw(app, win, "NIC :11:11:11", "network-tab")
    check_ip("Unknown")
    tab.find("IP address:", "push button").click()
    check_ip("10.0.0.3")

    win.window_close()
    app.topwin.grab_focus()

    # Tests the fake qemu guest agent path
    win = app.manager_open_details("test alternate devs title")
    tab = _select_hw(app, win, "NIC :11:72:72", "network-tab")
    check_ip("10.0.0.1", "fd00:beef::1/128")


def testDetailsEditDevices1(app):
    """
    Test all other devices
    """
    win = app.manager_open_details("test-many-devices")
    appl = win.find("config-apply", "push button")

    # Fail to hotremove
    tab = _select_hw(app, win, "Floppy 1", "disk-tab")
    tab.find("Advanced options", "toggle button").click_expander()
    share = tab.find("Shareable", "check box")
    share.click()
    lib.utils.check(lambda: appl.sensitive)
    win.find("config-remove").click()
    delete = app.find_window("Remove Disk")
    delete.find_fuzzy("Delete", "button").click()
    app.click_alert_button("change will take effect", "OK")
    lib.utils.check(lambda: not delete.showing)
    lib.utils.check(lambda: appl.sensitive)
    lib.utils.check(lambda: share.checked)
    win.find("config-cancel").click()

    _stop_vm(win)

    # Graphics simple VNC -> SPICE
    tab = _select_hw(app, win, "Display VNC", "graphics-tab")
    tab.combo_select("Type:", "Spice")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Spice GL example
    tab.combo_select("Listen type:", "None")
    tab.find("OpenGL:", "check box").click()
    tab.combo_check_default("graphics-rendernode", "0000")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Listen 'none' to default 'address' type triggers
    # a specific code path
    tab.combo_select("Listen type:", "Address")
    portauto = tab.find("graphics-port-auto", "check")
    lib.utils.check(lambda: portauto.visible)
    tab.find("OpenGL:", "check box")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    # Make sure it sticks
    tab.combo_check_default("Listen type:", "Address")

    # Switch to VNC with options
    tab.combo_select("Type:", "VNC")
    tab.combo_select("Listen type:", "Address")
    tab.find("graphics-port-auto", "check").click()
    tab.find("graphics-port", "spin button").set_text("6001")
    tab.find("Password:", "check").click()
    passwd = tab.find_fuzzy("graphics-password", "text")
    newpass = "foobar"
    passwd.typeText(newpass)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Sound device
    tab = _select_hw(app, win, "Sound sb16", "sound-tab")
    tab.find("Model:", "text").set_text("ac97")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    cell = win.find("Sound ac97", "table cell")
    oldtext = cell.text
    # Test non-disk removal
    win.find("config-remove").click()
    app.click_alert_button("Are you sure", "No")
    lib.utils.check(lambda: cell.state_selected)
    cell.click(button=3)
    app.root.find("Remove Hardware", "menu item").click()
    app.click_alert_button("Are you sure", "Yes")
    lib.utils.check(lambda: cell.text != oldtext)

    # Host device
    tab = _select_hw(app, win, "PCI 0000:00:19.0", "host-tab")
    tab.find("ROM BAR:", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Video device
    tab = _select_hw(app, win, "Video VMVGA", "video-tab")
    tab.find("Model:", "text").set_text("virtio")
    tab.find("3D acceleration:", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Watchdog
    tab = _select_hw(app, win, "Watchdog", "watchdog-tab")
    tab.find("Model:", "text").set_text("diag288")
    tab.find("Action:", "text").click()
    app.rawinput.pressKey("Down")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)


def testDetailsEditDevices2(app):
    app.uri = tests.utils.URIs.kvm_x86
    win = app.manager_open_details("test-many-devices", shutdown=True)
    appl = win.find("config-apply", "push button")

    # Controller SCSI
    tab = _select_hw(app, win, "Controller VirtIO SCSI 9", "controller-tab")
    tab.combo_select("controller-model", "Hypervisor")
    tab.find("SCSI Disk 1 on 9:0:0:0", "table cell")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Controller USB
    tab = _select_hw(app, win, "Controller USB 0", "controller-tab")
    tab.combo_select("controller-model", "USB 2")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab = _select_hw(app, win, "Controller USB 0", "controller-tab")
    tab.combo_select("controller-model", "USB 3")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)
    tab = _select_hw(app, win, "Controller USB 0", "controller-tab")
    tab.find("controller-model").find(None, "text").text = "piix3-uhci"
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Filesystem tweaks
    tab = _select_hw(app, win, "Filesystem /target/", "filesystem-tab")
    tab.combo_select("Driver:", "virtiofs")
    w = tab.find_fuzzy("Enable shared memory", "label")
    lib.utils.check(lambda: w.visible)
    tab.find("Source path:", "text").set_text("/frib1")
    tab.find("Target path:", "text").set_text("newtarget")
    tab.find_fuzzy("Export filesystem", "check box").click()
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # Smartcard tweaks
    tab = _select_hw(app, win, "Smartcard", "smartcard-tab")
    tab.combo_select("smartcard-mode", "Passthrough")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # TPM tweaks
    tab = _select_hw(app, win, "TPM", "tpm-tab")
    tab.combo_select("Type:", "Emulated")
    tab.find("Advanced options", "toggle button").click_expander()
    tab.combo_select("Model:", "CRB")
    tab.combo_select("Version:", "2.0")
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)

    # vsock tweaks
    tab = _select_hw(app, win, "VirtIO VSOCK", "vsock-tab")
    addr = tab.find("vsock-cid")
    auto = tab.find("vsock-auto")
    lib.utils.check(lambda: addr.text == "5")
    addr.set_text("7")
    appl.click()
    lib.utils.check(lambda: addr.text == "7")
    lib.utils.check(lambda: not appl.sensitive)
    auto.click()
    lib.utils.check(lambda: not addr.visible)
    appl.click()
    lib.utils.check(lambda: not appl.sensitive)


def testDetailsMiscEdits(app):
    """
    Test misc editing behavior, like checking for unapplied
    changes
    """
    app.open(extra_opts=["--test-options=test-update-device-fail"])
    win = app.manager_open_details("test-many-devices")
    hwlist = win.find("hw-list")

    # Live device removal, see results after shutdown
    disklabel = "SCSI Disk 1"
    tab = _select_hw(app, win, disklabel, "disk-tab")
    win.find("config-remove", "push button").click()
    delete = app.find_window("Remove Disk")
    delete.find_fuzzy("Delete", "button").click()

    # Will be fixed eventually
    app.click_alert_button("Device could not be removed", "OK")

    c = hwlist.find(disklabel, "table cell")
    _stop_vm(win)
    lib.utils.check(lambda: c.text != disklabel)

    # Remove a device for offline VM
    tab = _select_hw(app, win, "SCSI CDROM 1", "disk-tab")
    win.find("config-remove", "push button").click()
    delete = app.find_window("Remove Disk")
    delete.find_fuzzy("Delete", "button").click()
    lib.utils.check(lambda: win.active)

    # Attempt to apply changes when skipping away, but they fail
    tab.find("Advanced options", "toggle button").click_expander()
    tab.find("Cache mode:", "combo").find(None, "text").set_text("badcachemode")
    hwlist.find("CPUs", "table cell").click()
    app.click_alert_button("There are unapplied changes", "Yes")
    app.click_alert_button("badcachemode", "Close")

    # Cancelling changes
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    share = tab.find("Shareable:", "check box")
    lib.utils.check(lambda: not share.checked)
    share.click()
    win.find("config-cancel").click()
    lib.utils.check(lambda: not share.checked)

    # Unapplied, clicking no
    share = tab.find("Shareable:", "check box")
    share.click()
    hwlist.find("CPUs", "table cell").click()
    app.click_alert_button("There are unapplied changes", "No")
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    lib.utils.check(lambda: not share.checked)

    # Unapplied changes but clicking yes
    share.click()
    hwlist.find("CPUs", "table cell").click()
    alert = app.root.find("vmm dialog", "alert")
    alert.find_fuzzy("There are unapplied changes", "label")
    alert.find_fuzzy("Don't warn", "check box").click()
    alert.find("Yes", "push button").click()
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    lib.utils.check(lambda: share.checked)

    # Make sure no unapplied changes option sticks
    share.click()
    _select_hw(app, win, "CPUs", "cpu-tab")
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    lib.utils.check(lambda: share.checked)

    # VM State change doesn't refresh UI
    share.click()
    _start_vm(win)
    lib.utils.check(lambda: not share.checked)

    # Now apply changes to running VM, ensure they show up on shutdown
    win.find("config-apply").click()
    app.click_alert_button("changes will take effect", "OK")
    lib.utils.check(lambda: share.checked)
    _stop_vm(win)
    lib.utils.check(lambda: not share.checked)

    # Unapplied changes should warn when switching to XML tab
    tab = _select_hw(app, win, "Overview", "overview-tab")
    tab.find("Description:", "text").set_text("hey new description")
    win.find("XML", "page tab").click()
    # Select 'No', meaning don't abandon changes
    app.click_alert_button("changes will be lost", "No")
    lib.utils.check(lambda: tab.showing)

    # Try unapplied changes again, this time abandon our changes
    win.find("XML", "page tab").click()
    app.click_alert_button("changes will be lost", "Yes")
    lib.utils.check(lambda: not tab.showing)

    # Verify addhardware right click works
    win.find("Overview", "table cell").click(button=3)
    app.root.find("Add Hardware", "menu item").click()
    app.find_window("Add New Virtual Hardware")


def testDetailsXMLEdit(app):
    """
    Test XML editing interaction
    """
    app.open(xmleditor_enabled=True)
    win = app.manager_open_details("test-clone-simple")
    finish = win.find("config-apply")
    xmleditor = win.find("XML editor")

    # Edit vcpu count and verify it's reflected in CPU page
    tab = _select_hw(app, win, "CPUs", "cpu-tab")
    win.find("XML", "page tab").click()
    xmleditor.set_text(xmleditor.text.replace(">5</vcpu", ">8</vcpu"))
    finish.click()
    win.find("Details", "page tab").click()
    vcpualloc = tab.find("vCPU allocation:", "spin button")
    lib.utils.check(lambda: vcpualloc.text == "8")

    # Make some disk edits
    tab = _select_hw(app, win, "IDE Disk 1", "disk-tab")
    win.find("XML", "page tab").click()
    origpath = "/pool-dir/test-clone-simple.img"
    newpath = "/path/FOOBAR"
    xmleditor.set_text(xmleditor.text.replace(origpath, newpath))
    finish.click()
    win.find("Details", "page tab").click()
    disksrc = win.find("disk-source-path")
    lib.utils.check(lambda: disksrc.text == newpath)

    # Do standard xmleditor tests
    lib.utils.test_xmleditor_interactions(app, win, finish)


@unittest.mock.patch.dict("os.environ", {"VIRTINST_TEST_SUITE_FAKE_NO_SOURCEVIEW": "1"})
def testDetailsXMLEditorSourceviewFallback(app):
    """
    Test XML editor standard bits, when falling back to gtk textview
    """
    app.open(xmleditor_enabled=True)
    win = app.manager_open_details("test-clone-simple")
    finish = win.find("config-apply")

    # Do standard xmleditor tests
    lib.utils.test_xmleditor_interactions(app, win, finish)


def testDetailsConsoleChecksSSH(app):
    """
    Trigger a bunch of console connection failures to hit
    various details/* code paths
    """
    fakeuri = "qemu+ssh://foouser@256.256.256.256:1234/system"
    uri = tests.utils.URIs.test_full + ",fakeuri=%s" % fakeuri
    app.uri = uri
    app.open(xmleditor_enabled=True)

    app.topwin.find("test\n", "table cell").doubleClick()
    win = app.find_window("test on")
    conpages = win.find("console-pages")
    run = win.find("Run", "push button")
    shutdown = win.find("Shut Down", "push button")
    conbtn = win.find("Console", "radio button")
    detailsbtn = win.find("Details", "radio button")

    def _run():
        win.click_title()
        run.click()
        lib.utils.check(lambda: not run.sensitive)

    def _stop():
        shutdown.click()
        lib.utils.check(lambda: not shutdown.sensitive)

    def _checkcon(msg):
        conbtn.click()
        lib.utils.check(lambda: conpages.showing)
        conpages.find(msg)

    def _check_textconsole_menu(msg):
        vmenu = win.find("^View$", "menu")
        vmenu.click()
        tmenu = win.find("Consoles", "menu")
        tmenu.point()
        tmenu.find(msg, ".*menu item.*")
        vmenu.click()

    # Check initial state
    _checkcon("Graphical console not configured")
    _stop()
    _check_textconsole_menu("Graphical console not configured")

    # Add a SDL graphics device which can't be displayed
    detailsbtn.click()
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Graphics", "table cell").click()
    addhw.find("XML", "page tab").click()
    dev = '<graphics type="sdl" display=":3.4" xauth="/tmp/.Xauthority"/>'
    addhw.find("XML editor").text = dev
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: win.active)
    _run()
    _checkcon("Cannot display graphical console type")

    def _change_gfx_xml(_xml):
        detailsbtn.click()
        win.find("Display ", "table cell").click()
        win.find("XML", "page tab").click()
        win.find("XML editor").set_text(_xml)
        win.find("config-apply").click()

    # Listening from some other address
    _stop()
    xml = '<graphics type="spice" listen="0.0.0.0" port="6000" tlsPort="6001"/>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*resolving.*256.256.256.256.*")

    # Listening from some other address
    _stop()
    xml = '<graphics type="spice" listen="257.0.0.1" port="6000"/>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*resolving.*257.0.0.1.*")

    # Hit a specific error about tls only and ssh
    _stop()
    xml = '<graphics type="spice" tlsPort="60001" autoport="no"/>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*configured for TLS only.*")

    # Fake a socket connection
    _stop()
    xml = '<graphics type="vnc" socket="/tmp/foobar.sock"/>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*SSH tunnel error output.*")

    # Add a listen type='none' check
    _stop()
    xml = '<graphics type="spice"><listen type="none"/></graphics>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*local file descriptor.*")

    # Add a local list + port check
    _stop()
    xml = '<graphics type="spice" listen="127.0.0.1" port="6000" tlsPort="60001"/>'
    _change_gfx_xml(xml)
    _run()
    _checkcon(".*SSH tunnel error output.*")


def _testDetailsConsoleChecksTCP(app, fakeuri, msg):
    uri = tests.utils.URIs.test_full + ",fakeuri=%s" % fakeuri
    app.uri = uri
    app.open(xmleditor_enabled=True)

    app.topwin.find("test\n", "table cell").doubleClick()
    win = app.find_window("test on")
    conpages = win.find("console-pages")
    run = win.find("Run", "push button")
    shutdown = win.find("Shut Down", "push button")
    conbtn = win.find("Console", "radio button")
    detailsbtn = win.find("Details", "radio button")

    def _run():
        win.click_title()
        run.click()
        lib.utils.check(lambda: not run.sensitive)

    def _stop():
        shutdown.click()
        lib.utils.check(lambda: not shutdown.sensitive)

    def _checkcon(msg):
        conbtn.click()
        lib.utils.check(lambda: conpages.showing)
        conpages.find(msg)

    # Check initial state
    _checkcon("Graphical console not configured")
    _stop()

    detailsbtn.click()
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Graphics", "table cell").click()
    addhw.find("XML", "page tab").click()
    dev = '<graphics type="vnc" port="6000" address="127.0.0.1"/>'
    addhw.find("XML editor").text = dev
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: win.active)
    _run()
    _checkcon(msg)


def testDetailsConsoleTCPNonlocal(app):
    """
    Hit a specific warning when the connection has
    non-SSH transport but the guest config is only listening locally
    """
    fakeuri = "qemu+tcp://foouser@256.256.256.256:1234/system"
    msg = ".*configured to listen locally.*"
    _testDetailsConsoleChecksTCP(app, fakeuri, msg)


def testDetailsConsoleTCPLocal(app):
    """
    Ensure we don't hit the nonlocal check, when both graphics
    and connection URI are localhost
    """
    fakeuri = "qemu+tcp://foouser@localhost:1234/system"
    msg = "Viewer was disconnected\\."
    _testDetailsConsoleChecksTCP(app, fakeuri, msg)


def testDetailsConsoleSerialSwitch(app):
    """
    Test serial vs graphical console menu handling, across VM restarts,
    which can hit some corner cases.
    Hit a specific warning when the connection has
    non-SSH transport but the guest config is only listening locally
    """
    app.open(xmleditor_enabled=True)
    app.topwin.find("test\n", "table cell").doubleClick()
    win = app.find_window("test on")
    conpages = win.find("console-pages")
    run = win.find("Run", "push button")
    shutdown = win.find("Shut Down", "push button")
    conbtn = win.find("Console", "radio button")
    detailsbtn = win.find("Details", "radio button")

    def _run():
        win.click_title()
        run.click()
        lib.utils.check(lambda: not run.sensitive)

    def _stop():
        shutdown.click()
        lib.utils.check(lambda: not shutdown.sensitive)

    def _checkcon(msg):
        conbtn.click()
        lib.utils.check(lambda: conpages.showing)
        conpages.find(msg)

    def _open_textconsole_menu():
        conbtn.click()
        vmenu = win.find("^View$", "menu")
        vmenu.click()
        tmenu = win.find("Consoles", "menu")
        tmenu.point()
        app.sleep(0.5)  # give console menu time to dynamically populate
        return tmenu

    def _find_textconsole_item(msg):
        tmenu = _open_textconsole_menu()
        return tmenu.find(msg, "radio menu item")

    # Check initial state
    _checkcon("Graphical console not configured")
    _stop()

    # Add graphics device
    detailsbtn.click()
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Graphics", "table cell").click()
    addhw.find("XML", "page tab").click()
    dev = '<graphics type="spice" autoport="yes"><listen type="none"/></graphics>'
    addhw.find("XML editor").text = dev
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: win.active)

    # Add a serial
    # Add graphics device
    detailsbtn.click()
    win.find("add-hardware", "push button").click()
    addhw = app.find_window("Add New Virtual Hardware")
    addhw.find("Serial", "table cell").click()
    addhw.find("XML", "page tab").click()
    dev = '<serial type="pty"/>'
    addhw.find("XML editor").text = dev
    addhw.find("Finish", "push button").click()
    lib.utils.check(lambda: not addhw.active)
    lib.utils.check(lambda: win.active)

    _find_textconsole_item("Serial 1").click()
    _run()
    _checkcon(".*virDomainOpenConsole.*")
    _stop()
    _checkcon(".*Guest is not running.*")
    _run()
    _checkcon(".*virDomainOpenConsole.*")
    _stop()
    _checkcon(".*Guest is not running.*")
    sitem = _find_textconsole_item("Serial 1")
    lib.utils.check(lambda: sitem.sensitive)
    lib.utils.check(lambda: sitem.checked)

    # Remove serial
    detailsbtn.click()
    detailsbtn.click()
    _select_hw(app, win, "Serial 1", "char-tab")
    win.find("config-remove").click()
    app.click_alert_button("Are you sure", "Yes")
    _run()
    _checkcon(".*SPICE error-connect.*")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import unittest.mock

from . import lib


# UI tests for virt-manager's command line --show options


def testShowNewVM(app):
    app.open(uri="test:///default", extra_opts=["--show-domain-creator"])
    lib.utils.check(lambda: app.topwin.name == "New VM")
    app.topwin.window_close()
    app.wait_for_exit()


def testShowHost(app):
    app.open(uri="test:///default", extra_opts=["--show-host-summary"])

    lib.utils.check(lambda: app.topwin.name == "test default - Connection Details")
    nametext = app.topwin.find_fuzzy("Name:", "text")
    lib.utils.check(lambda: nametext.text == "test default")
    app.topwin.window_close()
    app.wait_for_exit()


def testShowDetails(app):
    app.open(extra_opts=["--show-domain-editor", "test-clone-simple"])

    lib.utils.check(lambda: "test-clone-simple on" in app.topwin.name)
    rlabel = app.topwin.find_fuzzy("Guest is not running", "label")
    lib.utils.check(lambda: not rlabel.showing)
    addhw = app.topwin.find_fuzzy("add-hardware", "button")
    lib.utils.check(lambda: addhw.showing)
    app.topwin.window_close()
    app.wait_for_exit()


def testShowPerformance(app):
    domid = "1"
    app.open(uri="test:///default", extra_opts=["--show-domain-performance", domid])

    lib.utils.check(lambda: "test on" in app.topwin.name)
    cpulabel = app.topwin.find_fuzzy("CPU usage", "label")
    lib.utils.check(lambda: cpulabel.showing)


def testShowConsole(app):
    # UUID of test-clone-simple
    uuid = "12345678-1234-ffff-1234-12345678ffff"
    app.open(extra_opts=["--show-domain-console", uuid])

    lib.utils.check(lambda: "test-clone-simple on" in app.topwin.name)
    rlabel = app.topwin.find_fuzzy("Guest is not running", "label")
    lib.utils.check(lambda: rlabel.showing)
    addhw = app.topwin.find_fuzzy("add-hardware", "button")
    lib.utils.check(lambda: not addhw.showing)


def testShowDelete(app):
    app.open(
        uri="test:///default", extra_opts=["--show-domain-delete", "test"], window_name="Delete"
    )
    # Ensure details opened too
    app.root.find("test on", "frame", check_active=False)

    delete = app.topwin
    delete.find_fuzzy("Delete", "button").click()
    app.wait_for_exit()


def testShowSystray(app):
    opts = ["--test-options=fake-systray", "--show-systray"]
    app.open(use_uri=False, extra_opts=opts, window_name="vmm-fake-systray")
    app.sleep(1)
    app.stop()

    app.open(uri="test:///default", extra_opts=opts, window_name="vmm-fake-systray")


def testShowRemoteDBusConnect(app):
    """
    Test the remote app dbus connection
    """
    app.open()
    lib.utils.check(lambda: "testdriver" in app.topwin.fmt_nodes())
    lib.utils.check(lambda: "test default" not in app.topwin.fmt_nodes())

    def _run_remote(opts):
        newapp = lib.app.VMMDogtailApp("test:///default")
        newapp.open(check_already_running=False, extra_opts=opts)
        timeout = 10
        lib.utils.check(lambda: not newapp.is_running(), timeout)
        vapps = [a for a in newapp.tree.root.applications() if a.name == "virt-manager"]
        lib.utils.check(lambda: len(vapps) == 1, timeout=timeout)
        # Ensure connection showed up
        app.topwin.find("test default", "table cell")

    _run_remote([])
    # Run remote again to trigger engine.py code when a connection
    # is already there and connected
    _run_remote(["--show-domain-console=test"])


def testShowCLIError(app):
    # Unknown option
    app.open(extra_opts=["--idontexist"])
    app.click_alert_button("Unhandled command line", "Close")
    lib.utils.check(lambda: not app.is_running())

    # Missing VM
    app.open(uri="test:///default", extra_opts=["--show-domain-delete", "IDONTEXIST"])
    app.click_alert_button("does not have VM", "Close")
    lib.utils.check(lambda: not app.is_running())

    # Bad URI
    baduri = "fribfrobfroo"
    app = lib.app.VMMDogtailApp(baduri)
    app.click_alert_button(baduri, "Close")
    lib.utils.check(lambda: not app.is_running())


def testCLIFirstRunURIGood(app):
    # Emulate first run with a URI that will succeed
    app.open(use_uri=False, firstrun_uri="test:///default")
    app.root.find("test default", "table cell")


def testCLIFirstRunURIBad(app):
    # Emulate first run with a URI that will not succeed
    app.open(use_uri=False, firstrun_uri="bad:///uri")
    app.topwin.find("bad uri", "table cell")
    app.click_alert_button("bad:///uri", "Close")


def testCLIFirstRunNoURI(app):
    # Emulate first run with no libvirtd detected
    app.open(use_uri=False, firstrun_uri="")
    errlabel = app.topwin.find("error-label")
    lib.utils.check(lambda: "Checking for virtualization" in errlabel.text)
    lib.utils.check(lambda: "detect a default hypervisor" in errlabel.text)


def testCLITraceLibvirt(app):
    # Just test this for code coverage
    app.open(
        keyfile="allstats.ini", extra_opts=["--trace-libvirt=mainloop", "--test-options=short-poll"]
    )
    app.sleep(0.5)  # Give time for polling to trigger
    lib.utils.check(lambda: app.topwin.active)


def testCLILeakDebug(app):
    # Just test this for code coverage
    app.open(
        keyfile="allstats.ini",
        extra_opts=["--test-options=leak-debug", "--test-options=short-poll"],
    )
    app.sleep(0.5)  # Give time for polling to trigger
    app.topwin.window_close()
    app.wait_for_exit()


def testCLINoFirstRun(app):
    # Test a simple case of loading without any config override
    app.open(first_run=False, enable_libguestfs=None, use_uri=False)
    lib.utils.check(lambda: app.topwin.showing)


def _testCLIFork(app, opts):
    app.open(
        first_run=False, enable_libguestfs=None, use_uri=False, allow_debug=False, extra_opts=opts
    )
    app.wait_for_exit()
    lib.utils.check(lambda: app.has_dbus())
    app.topwin.window_close()
    lib.utils.check(lambda: not app.has_dbus())


def testCLIFork(app):
    # Test app with --fork
    _testCLIFork(app, ["--fork"])


@unittest.mock.patch.dict("os.environ", {"VIRT_MANAGER_DEFAULT_FORK": "yes"})
def testCLIForkEnv(app):
    # Test with fork via env
    _testCLIFork(app, [])


def testCLIGTKArgs(app):
    # Ensure gtk arg passthrough works
    # Also test --no-fork is a no-op
    app.open(extra_opts=["--gtk-debug=misc", "--no-fork"])
    lib.utils.check(lambda: app.topwin.showing)


@unittest.mock.patch.dict("os.environ", {"DISPLAY": ""})
def testCLINoDisplay(app):
    # Ensure missing display exits
    app.open(will_fail=True)
    app.wait_for_exit()
